from __future__ import annotations

"""
OAuth2 Providers
"""

from .google import GoogleOAuth2Config, GoogleOAuth2Client
from .kinde import KindeOAuth2Config, KindeOAuth2Client
from typing import Dict, Union, Type, Optional, Tuple, TYPE_CHECKING
from lzl.load import lazy_import

ProviderConfigT = Union[
    GoogleOAuth2Config,
    KindeOAuth2Config,
]

ProviderClientT = Union[
    GoogleOAuth2Client,
    KindeOAuth2Client,
]

ProviderClasses: Dict[str , Tuple[Type[ProviderConfigT], Type[ProviderClientT]]] = {
    'google': (GoogleOAuth2Config, GoogleOAuth2Client),
    'kinde': (KindeOAuth2Config, KindeOAuth2Client),
}

if TYPE_CHECKING:
    from .google import TokenT as GoogleTokenT
    from .kinde import TokenT as KindeTokenT

    ProviderTokenT = Union[GoogleTokenT, KindeTokenT]

def get_provider_classes(
    name: str,
    config_class: Optional[str] = None,
    client_class: Optional[str] = None,
) -> Tuple[Type[ProviderConfigT], Type[ProviderClientT]]:
    """
    Returns the provider classes
    """
    if config_class is not None:
        if config_class in locals(): config_class = locals()[config_class]
        else: config_class = lazy_import(config_class)
    
    if client_class is not None:
        if client_class in locals(): client_class = locals()[client_class]
        else: client_class = lazy_import(client_class)
    
    if config_class and client_class:
        return config_class, client_class
    
    if name in ProviderClasses:
        return ProviderClasses[name][0], ProviderClasses[name][1]
    
    if 'google' in name:
        return ProviderClasses['google'][0], ProviderClasses['google'][1]
    
    if 'kinde' in name:
        return ProviderClasses['kinde'][0], ProviderClasses['kinde'][1]
    
    # if 'auth0' in name:
    #     return ProviderClasses['auth0'][0], ProviderClasses['auth0'][1]
    raise ValueError(f'Invalid Provider: {name}')


from __future__ import annotations

"""
The Hatchet Current Version

Patched: {{ timestamp }}
Version: {{ version }}
Last Version: {{ last_version }}
"""

from ..{{ current_version_name }} import Context, ContextT, Worker, CURRENT_VERSION, LAST_VERSION



from __future__ import annotations

"""
Postgres Configuration
"""

from ..base.config import (
    PostgresDsn,
    BaseDBConfig,
    BaseDBSettings,
    DatabaseConfig,
)
from lzl.types import eproperty
import typing as t

class BasePostgresConfig(BaseDBConfig):
    """
    The Base Postgres Config
    """
    url: t.Optional[PostgresDsn] = None
    readonly_url: t.Optional[PostgresDsn] = None
    superuser_url: t.Optional[PostgresDsn] = None

    @classmethod
    def get_adapter_prefix(cls) -> str:
        """
        Returns the adapter prefix
        """
        return 'postgresql'
    
    @eproperty
    def cli_connect_string(self) -> str:
        """
        Returns the cli connect string

        export PGPASSWORD=<password> psql -h <host> -p <port> -U <user> -d <database>
        """
        url = self.superuser_url or self.url
        url_data = url.hosts()[0]
        return f"export PGPASSWORD={url_data['password']} psql -h {url_data['host']} -p {url_data['port']} -U {url_data['username']} -d {url.path[1:]}"

    def get_cli_connect_string(
        self,
        superuser: t.Optional[bool] = None,
        readonly: t.Optional[bool] = None,
        database: t.Optional[str] = None,
        **kwargs,
    ) -> str:
        """
        Returns the cli connect string
        """
        if superuser: url = self.superuser_url
        elif readonly: url = self.readonly_url
        else: url = self.url
        assert url, 'URL must be provided'
        url_data = url.hosts()[0]
        db = database or url.path[1:]
        return f"export PGPASSWORD={url_data['password']} psql -h {url_data['host']} -p {url_data['port']} -U {url_data['username']} -d {db}"
    
    def get_cli_debug_args(self) -> str:
        """
        Returns the cli debug args
        """
        return ' -v ON_ERROR_STOP=1 -a -b -e -E'
    
    def get_cli_safe_cmd(self, cmd: str) -> str:
        """
        Returns the cli safe cmd
        """
        return cmd.split("psql", 1)[-1]


class PostgresSettings(BasePostgresConfig, BaseDBSettings):
    """
    The Postgres Settings
    """

    class Config:
        env_prefix = "POSTGRES_"
        case_sensitive = False


class PostgresConfig(BasePostgresConfig, DatabaseConfig):
    """
    The Postgres Config
    """
    pass






from __future__ import annotations

"""
OAuth2 Middleware

Unlike the OAuth Module, this has a few differences based on assumptions:

- If a user authenticates with the backend itself, it will generate an API Key 
  and store it in the session.

- Authenticating with the API Key assumes that the user has already authenticated via the backend.

- If Authentication is paseed as a header, it will assume that the user has already authenticated 
  via a 3rd party and will simply do a validation check of the Token.
"""


from .config import OAuth2Config
from .manager import OAuth2Manager
from .types.user import OAuth2User, UnauthenticatedUser, APIKeyUser
from .types.token import AccessToken, IDToken, OAuth2Token


from __future__ import annotations

"""Utility decorator for creating lazily initialised proxy objects."""

import typing as t

from .base import ProxyObject

ObjT = t.TypeVar("ObjT")
ProxyObjT = t.Type[ObjT]


@t.overload
def proxied(
    obj_cls: t.Optional[ObjT] = None,
    obj_getter: t.Optional[t.Union[t.Callable[..., ObjT], str]] = None,
    obj_args: t.Optional[t.List[t.Any]] = None,
    obj_kwargs: t.Optional[t.Dict[str, t.Any]] = None,
    obj_initialize: t.Optional[bool] = True,
    threadsafe: t.Optional[bool] = True,
) -> ObjT:
    ...


@t.overload
def proxied(
    **kwargs: t.Any,
) -> t.Callable[..., ObjT]:
    ...


def proxied(
    obj_cls: t.Optional[ObjT] = None,
    obj_getter: t.Optional[t.Union[t.Callable[..., ObjT], str]] = None,
    obj_args: t.Optional[t.List[t.Any]] = None,
    obj_kwargs: t.Optional[t.Dict[str, t.Any]] = None,
    obj_initialize: t.Optional[bool] = True,
    threadsafe: t.Optional[bool] = True,
) -> t.Union[t.Callable[..., ObjT], ObjT]:
    """Return a proxy that defers constructing ``obj_cls`` until first use."""

    if obj_cls is not None:
        return ProxyObject(
            obj_cls=obj_cls,
            obj_getter=obj_getter,
            obj_args=obj_args,
            obj_kwargs=obj_kwargs,
            obj_initialize=obj_initialize,
            threadsafe=threadsafe,
        )

    def wrapper(inner_cls: ObjT) -> ProxyObjT:
        return ProxyObject(
            obj_cls=inner_cls,
            obj_getter=obj_getter,
            obj_args=obj_args,
            obj_kwargs=obj_kwargs,
            obj_initialize=obj_initialize,
            threadsafe=threadsafe,
        )

    return wrapper


__all__ = ["proxied", "ObjT", "ProxyObjT"]



from __future__ import annotations

"""Convenience exports for LazyOps lazy-loading utilities.

The package provides a small façade around :class:`LazyLoad` and supporting
helpers so downstream modules can defer importing heavy dependencies until
absolutely necessary.  Behaviour matches the original implementation, but the
public interface is now documented for automated reference generation.
"""

import typing as t

from .main import LazyLoad, lazy_load, load
from .wrappers import lazy_function_wrapper
from .utils import (
    import_from_string,
    import_function,
    import_string,
    is_coro_func,
    lazy_function,
    lazy_import,
    validate_callable,
)

__all__ = [
    "LazyLoad",
    "lazy_load",
    "load",
    "lazy_function_wrapper",
    "lazy_import",
    "lazy_function",
    "is_coro_func",
    "import_string",
    "import_function",
    "import_from_string",
    "validate_callable",
    "TYPE_CHECKING",
]

TYPE_CHECKING = t.TYPE_CHECKING



from __future__ import annotations

"""
Kubernetes Leader Election for FastAPI Applications

This module provides a leader election mechanism for Kubernetes deployments,
ensuring that only one pod/worker handles specific processes at a time.
"""

import os
import time
import uuid
import asyncio
from contextlib import asynccontextmanager
from datetime import datetime, timedelta, timezone
from typing import Optional, Callable, Any, TYPE_CHECKING
from functools import wraps
from lzl.logging import logger
from lzl import load

if TYPE_CHECKING:
    import kubernetes
    import kubernetes.client.rest
    import kubernetes.client.exceptions
    import fastapi
    # from kubernetes import client, config
    # from kubernetes.client.rest import ApiException

else:
    kubernetes = load.lazy_load('kubernetes')
    fastapi = load.lazy_load('fastapi')



class KubernetesLeaderElection:
    """
    Implements leader election using Kubernetes Lease objects.
    
    This class ensures that only one instance (pod/worker) can hold the leadership
    role at any given time, with automatic failover if the leader fails.
    """
    
    def __init__(
        self,
        lease_name: str,
        namespace: str = None,
        lease_duration_seconds: int = 15,
        renew_deadline_seconds: int = 10,
        retry_period_seconds: int = 2,
        identity: str = None
    ):
        """
        Initialize the leader election manager.
        
        Args:
            lease_name: Name of the Kubernetes Lease object
            namespace: Kubernetes namespace (defaults to current pod's namespace)
            lease_duration_seconds: How long the lease is valid
            renew_deadline_seconds: How long before lease expires to renew
            retry_period_seconds: How often to retry acquiring the lease
            identity: Unique identifier for this instance (defaults to pod name + worker id)
        """
        self.lease_name = lease_name
        self.namespace = namespace or self._get_namespace()
        self.lease_duration_seconds = lease_duration_seconds
        self.renew_deadline_seconds = renew_deadline_seconds
        self.retry_period_seconds = retry_period_seconds
        self.identity = identity or self._generate_identity()
        
        self.is_leader = False
        self._stop_election = False
        self._election_task = None
        
        # Initialize Kubernetes client
        try:
            kubernetes.config.load_incluster_config()
        except kubernetes.config.ConfigException:
            # Fallback for local development
            kubernetes.config.load_kube_config()

        self.coordination_v1 = kubernetes.client.CoordinationV1Api()
    
    def _get_namespace(self) -> str:
        """Get the current pod's namespace."""
        namespace_path = "/var/run/secrets/kubernetes.io/serviceaccount/namespace"
        if os.path.exists(namespace_path):
            with open(namespace_path, 'r') as f:
                return f.read().strip()
        return "default"
    
    def _generate_identity(self) -> str:
        """Generate a unique identity for this instance."""
        from lzo.utils.system import get_host_name
        pod_name = os.environ.get("HOSTNAME", f"{get_host_name()}-{uuid.uuid4().hex[:8]}")
        worker_id = os.environ.get("WORKER_ID", str(os.getpid()))
        return f"{pod_name}-{worker_id}"

    def _create_or_get_lease(self) -> Optional['kubernetes.client.V1Lease']:
        """Create or retrieve the lease object."""
        try:
            return self.coordination_v1.read_namespaced_lease(
                name=self.lease_name,
                namespace=self.namespace
            )
        except kubernetes.client.rest.ApiException as e:
            if e.status == 404:
                # Lease doesn't exist, create it
                now = datetime.now(timezone.utc)
                lease = kubernetes.client.V1Lease(
                    metadata=kubernetes.client.V1ObjectMeta(
                        name=self.lease_name,
                        namespace=self.namespace
                    ),
                    spec=kubernetes.client.V1LeaseSpec(
                        holder_identity=self.identity,
                        lease_duration_seconds=self.lease_duration_seconds,
                        acquire_time=now,
                        renew_time=now
                    )
                )
                try:
                    return self.coordination_v1.create_namespaced_lease(
                        namespace=self.namespace,
                        body=lease
                    )
                except kubernetes.client.rest.ApiException as create_error:
                    if create_error.status == 409:
                        # Another instance created it concurrently
                        return self.coordination_v1.read_namespaced_lease(
                            name=self.lease_name,
                            namespace=self.namespace
                        )
                    raise
            raise


    def _parse_k8s_datetime(self, dt_value) -> datetime:
        """Parse datetime from Kubernetes, handling both string and datetime objects."""
        if dt_value is None:
            return None
        
        if isinstance(dt_value, datetime):
            # Already a datetime object
            if dt_value.tzinfo is None:
                # Assume UTC if no timezone
                return dt_value.replace(tzinfo=timezone.utc)
            return dt_value
        
        if isinstance(dt_value, str):
            # Handle various datetime string formats from Kubernetes
            # Remove microseconds if present (everything after the dot before timezone)
            if '.' in dt_value:
                # Split at the dot and reconstruct without microseconds
                base, remainder = dt_value.split('.', 1)
                # Find where timezone info starts (Z, +, or -)
                for i, char in enumerate(remainder):
                    if char in ['Z', '+', '-']:
                        dt_value = base + remainder[i:]
                        break
                else:
                    dt_value = base  # No timezone found
            
            # Replace Z with +00:00 for ISO format compatibility
            dt_value = dt_value.replace('Z', '+00:00')
            
            try:
                return datetime.fromisoformat(dt_value)
            except ValueError:
                # Fallback to parsing without timezone and assume UTC
                try:
                    dt_clean = dt_value.split('+')[0].split('-')[0] if '+' in dt_value or dt_value.count('-') > 2 else dt_value
                    return datetime.fromisoformat(dt_clean).replace(tzinfo=timezone.utc)
                except Exception as e:
                    logger.warning(f"Could not parse datetime: {dt_value}, using current time: {e}")
                    return datetime.now(timezone.utc)
        
        # Fallback for unexpected types
        logger.warning(f"Unexpected datetime type: {type(dt_value)}, using current time")
        return datetime.now(timezone.utc)
    
    def _try_acquire_or_renew(self) -> bool:
        # sourcery skip: extract-duplicate-method, hoist-statement-from-if, remove-unnecessary-else, swap-if-else-branches
        """Try to acquire or renew the lease."""
        lease = self._create_or_get_lease()
        if not lease:
            return False
        
        now = datetime.now(timezone.utc)
        
        # Check if we can acquire the lease
        if lease.spec.holder_identity == self.identity:
            # We already hold the lease, renew it
            lease.spec.renew_time = now
        elif lease.spec.renew_time:
            # Check if the current lease has expired
            renew_time = self._parse_k8s_datetime(lease.spec.renew_time)
            expiry_time = renew_time + timedelta(seconds=self.lease_duration_seconds)
            
            if now > expiry_time:
                # Lease has expired, we can acquire it
                lease.spec.holder_identity = self.identity
                lease.spec.acquire_time = now
                lease.spec.renew_time = now
            else:
                # Lease is held by another instance
                return False
        else:
            # No renew time set, acquire the lease
            lease.spec.holder_identity = self.identity
            lease.spec.acquire_time = now
            lease.spec.renew_time = now
        
        # Try to update the lease
        try:
            self.coordination_v1.replace_namespaced_lease(
                name=self.lease_name,
                namespace=self.namespace,
                body=lease
            )
            return lease.spec.holder_identity == self.identity
        except kubernetes.client.rest.ApiException as e:
            if e.status == 409:
                # Conflict - another instance updated the lease
                return False
            raise
    
    async def _election_loop(self):  # sourcery skip: remove-redundant-if
        """Main election loop that runs continuously."""
        retry_count = 0
        max_retries = 5
        
        while not self._stop_election:
            try:
                acquired = self._try_acquire_or_renew()
                retry_count = 0  # Reset retry count on success
                
                if acquired and not self.is_leader:
                    logger.info(f"Leadership acquired: |g|{self.identity}|e|", colored = True)
                    self.is_leader = True
                    if hasattr(self, '_on_elected_callback'):
                        await self._on_elected_callback()
                elif not acquired and self.is_leader:
                    logger.info(f"Leadership lost: |y|{self.identity}|e|", colored = True)
                    self.is_leader = False
                    if hasattr(self, '_on_lost_callback'):
                        await self._on_lost_callback()
                elif acquired and self.is_leader:
                    logger.debug(f"Leadership renewed: |g|{self.identity}|e|", colored = True)

                # Sleep before next attempt
                if self.is_leader:
                    # If we're the leader, renew more frequently
                    await asyncio.sleep(self.renew_deadline_seconds)
                else:
                    # If we're not the leader, check less frequently
                    await asyncio.sleep(self.retry_period_seconds)
                    
            except kubernetes.client.rest.ApiException as e:
                retry_count += 1
                logger.error(f"Kubernetes API error in election loop (attempt {retry_count}/{max_retries}): {e.status} - {e.reason}")
                
                if e.status == 401:
                    logger.error("Authentication failed. Check service account permissions.")
                elif e.status == 403:
                    logger.error("Authorization failed. Check RBAC permissions for leases.")
                
                if retry_count >= max_retries:
                    logger.error(f"Max retries ({max_retries}) reached. Giving up leadership.")
                    self.is_leader = False
                    retry_count = 0  # Reset for next cycle
                
                await asyncio.sleep(self.retry_period_seconds * retry_count)  # Exponential backoff
                
            except Exception as e:
                retry_count += 1
                logger.error(f"Unexpected error in election loop (attempt {retry_count}/{max_retries}): {e}", exc_info=True)
                
                if retry_count >= max_retries:
                    logger.error(f"Max retries ({max_retries}) reached. Giving up leadership.")
                    self.is_leader = False
                    retry_count = 0
                
                await asyncio.sleep(self.retry_period_seconds * retry_count)
    
    async def start(
        self,
        on_elected: Optional[Callable] = None,
        on_lost: Optional[Callable] = None
    ):
        """
        Start the leader election process.
        
        Args:
            on_elected: Async callback when leadership is acquired
            on_lost: Async callback when leadership is lost
        """
        if on_elected:
            self._on_elected_callback = on_elected
        if on_lost:
            self._on_lost_callback = on_lost
        
        self._stop_election = False
        self._election_task = asyncio.create_task(self._election_loop())
        logger.info(f"Leader election started for {self.identity}"
                    )
    
    async def stop(self):
        """Stop the leader election process."""
        self._stop_election = True
        if self._election_task:
            await self._election_task
        self.is_leader = False
        logger.info(f"Leader election stopped for {self.identity}")

    def leader_only(self, func: Callable) -> Callable:
        """
        Decorator to ensure a function only runs on the leader.
        
        Usage:
            @leader_election.leader_only
            async def process_batch():
                # This will only run on the leader
                pass
        """
        @wraps(func)
        async def wrapper(*args, **kwargs):
            if not self.is_leader:
                raise fastapi.HTTPException(
                    status_code=503,
                    detail="This instance is not the leader"
                )
            return await func(*args, **kwargs)
        return wrapper
    
    @asynccontextmanager
    async def as_leader(self):
        """
        Context manager that only executes if this instance is the leader.
        
        Usage:
            async with leader_election.as_leader():
                # Code here only runs on the leader
                await process_important_task()
        """
        if self.is_leader:
            yield
        else:
            raise fastapi.HTTPException(
                status_code=503,
                detail="This instance is not the leader"
            )


# # FastAPI integration example
# def create_leader_election_app(
#     lease_name: str = "my-app-leader",
#     namespace: str = None
# ) -> tuple[FastAPI, KubernetesLeaderElection]:
#     """
#     Create a FastAPI app with leader election integrated.
    
#     Returns:
#         Tuple of (FastAPI app, KubernetesLeaderElection instance)
#     """
    
#     leader_election = KubernetesLeaderElection(
#         lease_name=lease_name,
#         namespace=namespace
#     )
    
#     @asynccontextmanager
#     async def lifespan(app: FastAPI):
#         # Startup
#         async def on_elected():
#             logger.info("This instance is now the leader!")
#             # Initialize leader-only resources here
        
#         async def on_lost():
#             logger.info("This instance lost leadership!")
#             # Cleanup leader-only resources here
        
#         await leader_election.start(
#             on_elected=on_elected,
#             on_lost=on_lost
#         )
        
#         yield
        
#         # Shutdown
#         await leader_election.stop()
    
#     app = FastAPI(lifespan=lifespan)
    
#     # Health check endpoint
#     @app.get("/health")
#     async def health():
#         return {
#             "status": "healthy",
#             "is_leader": leader_election.is_leader,
#             "identity": leader_election.identity
#         }
    
#     # Leader status endpoint
#     @app.get("/leader")
#     async def leader_status():
#         return {
#             "is_leader": leader_election.is_leader,
#             "identity": leader_election.identity,
#             "lease_name": leader_election.lease_name
#         }
    
#     # Example endpoint that only works on the leader
#     @app.post("/process-batch")
#     @leader_election.leader_only
#     async def process_batch(data: dict):
#         """This endpoint only processes on the leader instance."""
#         # Your batch processing logic here
#         return {
#             "status": "processed",
#             "leader": leader_election.identity,
#             "data": data
#         }
    
#     return app, leader_election




from __future__ import annotations

"""Registry of singleton-like objects for lazy loading.

.. important::
   This module mirrors legacy behaviour and is intentionally left mostly
   untouched.  Several sections are incomplete (see
   ``docs/future-updates.md``) and should be revisited in a future refactor.
"""

import typing as t

if t.TYPE_CHECKING:
    from .mixins import RegisteredObject

RObjectT = t.TypeVar('RObjectT', bound='RegisteredObject')

_module_object_registry: t.Dict[str, t.Type['RObjectT']] = {}
_init_module_object_registry: t.Dict[str, 'RObjectT'] = {}
_uninit_module_object_registry: t.Dict[str, t.Dict[str, str]] = {}


def register_object(obj: t.Type['RObjectT']) -> None:
    """Register a lazily loadable object.

    Note:
        The implementation mirrors the legacy version and still contains
        unresolved TODOs.
    """

    global _module_object_registry, _uninit_module_object_registry
    from lzl.io.ser import get_object_classname
    from lzl.logging import logger

    if hasattr(obj, '_rxtra') and obj._rxtra.get('registered'):
        logger.warning(
            'Object %s already registered with %s',
            obj._rxtra['obj_ref'],
            obj._rxtra['module'],
        )
        return

    cls_name = get_object_classname(obj, is_type=True)
    cls_module = obj.__module__.split('.')[0]

    obj_ref = cls_module
    if hasattr(obj, 'kind'):
        obj_ref += f'.{obj.kind}'
    if hasattr(obj, 'name'):
        obj_ref += f'.{obj.name}'

    if cls_module not in _module_object_registry:
        _module_object_registry[cls_module] = {}

    global _module_client_registry, _uninit_module_client_registry  # type: ignore[name-defined]
    from lzl.io.ser import get_object_classname
    from lzl.logging import logger

    cls_name = get_object_classname(client, is_type=True)  # type: ignore[name-defined]
    cls_module = client.__module__.split('.')[0]  # type: ignore[name-defined]

    if cls_module not in _module_client_registry:
        _module_client_registry[cls_module] = {}



from __future__ import annotations

"""
Helpers
"""
import json
import yaml
import base64
from pathlib import Path
from typing import Dict, Any, Optional, List, TYPE_CHECKING

def to_yaml(data, indent: Optional[int] = None):
    return yaml.safe_dump(data, default_flow_style = False, indent = indent)

def to_json(data, indent: Optional[int] = None):
    return json.dumps(data, indent = indent)

def to_base64(data: Dict[str, str]):
    """
    Converts a dict to a base64 string
    """
    return {k: base64.b64encode(v.encode('utf-8')).decode('utf-8') for k, v in data.items()}

def get_http(url: str, headers: Optional[Dict[str, str]] = None, timeout: Optional[int] = None, **kwargs) -> str:
    """
    Returns the HTTP Response
    """
    from lzl.api import aiohttpx
    return aiohttpx.get(url, headers = headers, timeout = timeout, **kwargs).text

def load_file(path: str) -> str:
    """
    Loads a file
    """
    from lzl.io.file import File
    return File(path).read_text()

def to_title(text: str) -> str:
    """
    Converts a string to title case
    """
    return text.replace('_', ' ').title()

def normalize_value(value: str | list | Any) -> Any:
    """
    Normalizes a value
    """
    if isinstance(value, list): return ', '.join(value)
    return value.strip() if isinstance(value, str) else value


def normalize_ids_to_str(ids: List[str | int]) -> str:
    """
    Normalizes the ids to a string
    """
    if isinstance(ids[0], int): return ', '.join([str(id) for id in ids])
    return ', '.join([f"'{id}'" for id in ids])

autofilters = {
    'to_json': to_json,
    'to_yaml': to_yaml,
    'to_base64': to_base64,
    'get_http': get_http,
    'load_file': load_file,
    'to_title': to_title,
    'normalize_value': normalize_value,
    'normalize_ids_to_str': normalize_ids_to_str,
}


from __future__ import annotations

"""
SQLite Backend

Borrowed parts from `python-diskache <https://github.com/grantjenks/python-diskcache>`_
"""

import abc
import copy
import time
import asyncio
import contextlib
import threading
import contextlib
import typing as t
from lzl import load
from lzl.pool import ThreadPool
from lzo.types import eproperty
from pydantic.networks import UrlConstraints, Annotated
try:
    from pydantic_core import MultiHostUrl
except ImportError:
    from pydantic.networks import MultiHostUrl
from ..base import logger
from .utils import dict_diff
from typing import TypeVar, Generic, Any, Dict, Optional, Union, Tuple, Iterable, List, Type, Callable, Generator, AsyncGenerator, TYPE_CHECKING

if load.TYPE_CHECKING:
    import sqlite3
    import aiosqlite

    from sqlite3 import Connection, Cursor
    from aiosqlite import Connection as AsyncConnection, Cursor as AsyncCursor
else:
    sqlite3 = load.LazyLoad("sqlite3")
    aiosqlite = load.LazyLoad("aiosqlite")


from .base import (
    BaseSqliteDB,
    get_eviction_policies,
    get_optimized_settings,
    METADATA,
    OPTIMIZED,
    SqliteDsn,
    Optimization,

    Timeout,
    ENOVAL,
    UNKNOWN,
)

class SqliteDB(BaseSqliteDB):
    """
    The Sqlite Database
    """

    def get_child(self, table: str, timeout: t.Optional[float] = None, **kwargs) -> 'SqliteDB':
        """
        Creates a New Child Instance
        """
        child_kwargs = copy.deepcopy(self._child_kwargs)
        child_kwargs['table'] = table
        if timeout is not None: child_kwargs['timeout'] = timeout
        if self._debug_enabled: child_kwargs['debug_enabled'] = self._debug_enabled
        if kwargs: child_kwargs['kwargs'].update(kwargs)
        child_kwargs['is_child_db'] = True
        kws = child_kwargs.pop('kwargs', {})
        # logger.info({**child_kwargs, **kws}, prefix = f'{table} - Init Kws', colored = True)
        new_child = type(self)(**child_kwargs, **kws)
        # Provide the underlying props
        _extra_kws = copy.deepcopy(self._extra)
        _extra_kws.pop('is_configured', None)
        _extra_kws.pop('configured', None)
        _extra_kws.pop('has_configured_table', None)
        _extra_kws.pop('has_configured_settings', None)
        _extra_kws.pop('eviction_policies', None)
        _extra_kws.pop('eviction_policy', None)
        new_child._extra.update(_extra_kws)
        # logger.info(f'Child Cache: {new_child.conn_uri}: {new_child.table}: {new_child._extra}')
        # logger.info(new_child._extra, prefix = f'{new_child.table} - Extra', colored = True)
        return new_child



    """
    Public Primary Methods
    """
    
    def _set(self, key: str, value: t.Any, expire: Optional[float] = None, read: bool = False, tag: t.Optional[str] = None, retry: bool = False):
        """Set `key` and `value` item in cache.

        When `read` is `True`, `value` should be a file-like object opened
        for reading in binary mode.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key for item
        :param value: value for item
        :param float expire: seconds until item expires
            (default None, no expiry)
        :param bool read: read value as bytes from file (default False)
        :param str tag: text to associate with key (default None)
        :param bool retry: retry if database timeout occurs (default False)
        :return: True if item was set
        :raises Timeout: if database timeout occurs
        """
        now = time.time()
        size = self.get_object_size(value)
        expire_time = None if expire is None else now + expire
        columns = (expire_time, tag, size, value)

        # The order of SELECT, UPDATE, and INSERT is important below.
        #
        # Typical cache usage pattern is:
        #
        # value = cache.get(key)
        # if value is None:
        #     value = expensive_calculation()
        #     cache.set(key, value)
        #
        # Cache.get does not evict expired keys to avoid writes during lookups.
        # Commonly used/expired keys will therefore remain in the cache making
        # an UPDATE the preferred path.
        #
        # The alternative is to assume the key is not present by first trying
        # to INSERT and then handling the IntegrityError that occurs from
        # violating the UNIQUE constraint. This optimistic approach was
        # rejected based on the common cache usage pattern.
        #
        # INSERT OR REPLACE aka UPSERT is not used because the old filename may
        # need cleanup.

        with self._transact(retry) as sql:
            rows = sql(
                f'SELECT rowid FROM "{self.table}"'
                ' WHERE key = ?',
                (key,),
            ).fetchall()

            if rows:
                ((rowid,),) = rows
                # logger.info(f'{rowid}') 
                self._row_update(rowid, now, columns)
            else:
                self._row_insert(key, now, columns)

            self._cull(now, sql)
            # self._cull(now, sql, cleanup)
            return True

    def set(self, key: str, value: t.Any, expire: Optional[float] = None, read: bool = False, tag: t.Optional[str] = None, retry: bool = False):
        """Set `key` and `value` item in cache.

        When `read` is `True`, `value` should be a file-like object opened
        for reading in binary mode.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key for item
        :param value: value for item
        :param float expire: seconds until item expires
            (default None, no expiry)
        :param bool read: read value as bytes from file (default False)
        :param str tag: text to associate with key (default None)
        :param bool retry: retry if database timeout occurs (default False)
        :return: True if item was set
        :raises Timeout: if database timeout occurs
        """
        now = time.time()
        size = self.get_object_size(value)
        expire_time = None if expire is None else now + expire
        columns = (expire_time, tag, size, value)
        
        with self._transact(retry) as sql:
            # We'll use the upsert method here
            self._row_upsert(key, now, columns, sql = sql)
            self._cull(now, sql)
            return True
        
    async def _aset(self, key: str, value: t.Any, expire: Optional[float] = None, read: bool = False, tag: t.Optional[str] = None, retry: bool = False):
        """Set `key` and `value` item in cache.

        When `read` is `True`, `value` should be a file-like object opened
        for reading in binary mode.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key for item
        :param value: value for item
        :param float expire: seconds until item expires
            (default None, no expiry)
        :param bool read: read value as bytes from file (default False)
        :param str tag: text to associate with key (default None)
        :param bool retry: retry if database timeout occurs (default False)
        :return: True if item was set
        :raises Timeout: if database timeout occurs
        """
        now = time.time()
        size = self.get_object_size(value)
        expire_time = None if expire is None else now + expire
        columns = (expire_time, tag, size, value)        
        async with self._atransact(retry) as sql:
            # result = await sql(
            #     f'SELECT rowid FROM {self.table}'
            #     ' WHERE key = ?',
            #     (key,),
            # )
            # rows = await result.fetchall()
            rows = await (await sql(
                f'SELECT rowid FROM "{self.table}"'
                ' WHERE key = ?',
                (key,),
            )).fetchall()

            if rows:
                ((rowid,),) = rows
                await self._row_update(rowid, now, columns, is_async = True)
            else:
                await self._row_insert(key, now, columns, is_async = True)

            self.abg_cull(now, retry = retry)
            # await self._acull(now, sql)
            return True
    

    async def aset(self, key: str, value: t.Any, expire: Optional[float] = None, read: bool = False, tag: t.Optional[str] = None, retry: bool = False):
        """Set `key` and `value` item in cache.

        When `read` is `True`, `value` should be a file-like object opened
        for reading in binary mode.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key for item
        :param value: value for item
        :param float expire: seconds until item expires
            (default None, no expiry)
        :param bool read: read value as bytes from file (default False)
        :param str tag: text to associate with key (default None)
        :param bool retry: retry if database timeout occurs (default False)
        :return: True if item was set
        :raises Timeout: if database timeout occurs
        """
        now = time.time()
        size = self.get_object_size(value)
        expire_time = None if expire is None else now + expire
        columns = (expire_time, tag, size, value)        
        async with self._atransact(retry) as sql:
            await self._row_upsert(key, now, columns, is_async = True, sql = sql)
            self.abg_cull(now, retry = retry)
            return True

    def batch_set(self, data: t.Dict[str, t.Any], expire: t.Optional[float] = None, tag: t.Optional[str] = None, retry: t.Optional[bool] = None):
        """
        Batch sets many items

        :param data: dict of key/value pairs
        :param expire: seconds until the key expires
            (default None, no expiry)
        :param retry: retry if database timeout occurs (default None, no retry)
        :return: True if item was set
        """
        now = time.time()
        expire_time = None if expire is None else now + expire
        columns = (expire_time, tag)
        with self._transact(retry) as sql:
            self._bulk_upsert(data, now, columns)
            self._cull(now, sql)
            return True

    async def abatch_set(self, data: t.Dict[str, t.Any], expire: t.Optional[float] = None, tag: t.Optional[str] = None, retry: t.Optional[bool] = None):
        """
        Batch sets many items

        :param data: dict of key/value pairs
        :param expire: seconds until the key expires
            (default None, no expiry)
        :param retry: retry if database timeout occurs (default None, no retry)
        :return: True if item was set
        """
        now = time.time()
        expire_time = None if expire is None else now + expire
        columns = (expire_time, tag)
        # async with self._atransact(retry) as sql:
        await self._bulk_upsert(data, now, columns, is_async = True)
        self.abg_cull(now, retry = retry)
        return True


    def _cull(self, now: float, sql: t.Callable[..., 'Cursor'], limit: t.Optional[int] = None):
        """
        Culls the cache
        """
        cull_limit = self.cull_limit if limit is None else limit
        if cull_limit == 0: return

        # Evict expired keys.
        select_expired_template = (
            f'SELECT %s FROM "{self.table}"'
            ' WHERE expire_time IS NOT NULL AND expire_time < ?'
            ' ORDER BY expire_time LIMIT ?'
        )

        select_expired = select_expired_template % 'key'
        rows = sql(select_expired, (now, cull_limit)).fetchall()
        if rows:
            delete_expired = f'DELETE FROM "{self.table}" WHERE rowid IN (%s)' % (
                select_expired_template % 'rowid'
            )
            sql(delete_expired, (now, cull_limit))
            cull_limit -= len(rows)
            if cull_limit == 0: return

        # Evict keys by policy.
        select_policy = self.eviction_policies['cull']
        if select_policy is None or self.volume() < self.size_limit:
            return
        
        select_key = select_policy.format(fields='key', now=now)
        rows = sql(select_key, (cull_limit,)).fetchall()
        if rows:
            delete = f'DELETE FROM "{self.table}" WHERE key IN (%s)' % (
                select_policy.format(fields='rowid', now=now)
            )
            sql(delete, (cull_limit,))
    

    async def _acull(self, now: float, sql: t.Callable[..., t.Awaitable['AsyncCursor']], limit: t.Optional[int] = None):
        """
        Culls the cache
        """
        cull_limit = self.cull_limit if limit is None else limit
        if cull_limit == 0: return

        # Evict expired keys.
        select_expired_template = (
            f'SELECT %s FROM "{self.table}"'
            ' WHERE expire_time IS NOT NULL AND expire_time < ?'
            ' ORDER BY expire_time LIMIT ?'
        )

        select_expired = select_expired_template % 'key'
        rows = await (await sql(select_expired, (now, cull_limit))).fetchall()
        if rows:
            delete_expired = f'DELETE FROM "{self.table}" WHERE rowid IN (%s)' % (
                select_expired_template % 'rowid'
            )
            await sql(delete_expired, (now, cull_limit))
            cull_limit -= len(rows)
            if cull_limit == 0: return

        # Evict keys by policy.
        select_policy = self.eviction_policies['cull']
        if select_policy is None or (await self.avolume()) < self.size_limit:
            return
        
        select_key = select_policy.format(fields='key', now=now)
        rows = (await sql(select_key, (cull_limit,))).fetchall()
        if rows:
            delete = f'DELETE FROM "{self.table}" WHERE key IN (%s)' % (
                select_policy.format(fields='rowid', now=now)
            )
            await sql(delete, (cull_limit,))

    async def _abg_cull(self, now: float, limit: t.Optional[int] = None, retry: bool = False):
        """
        [Async - Background] Culls the cache
        """
        # We don't pass the sql connection for thread safety.
        # async with self._atransact(retry) as sql:
        sql = self._asql_retry if retry else self._asql
        await self._acull(now, sql, limit = limit)
    
    def abg_cull(self, now: float, limit: t.Optional[int] = None, retry: bool = False):
        """
        [Async Background] Culls the cache
        """
        return self.pool.create_background(self._abg_cull, now = now, limit = limit, retry = retry)

    def touch(self, key: str, expire: Optional[float] = None, retry: bool = False):
        """Touch `key` in cache and update `expire` time.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key for item
        :param float expire: seconds until item expires
            (default None, no expiry)
        :param bool retry: retry if database timeout occurs (default False)
        :return: True if key was touched
        :raises Timeout: if database timeout occurs
        """
        now = time.time()
        expire_time = None if expire is None else now + expire
        with self._transact(retry) as sql:
            rows = sql(
                f'SELECT rowid, expire_time FROM "{self.table}"'
                ' WHERE key = ?',
                (key,),
            ).fetchall()
            if rows:
                ((rowid, old_expire_time),) = rows

                if old_expire_time is None or old_expire_time > now:
                    sql(
                        f'UPDATE "{self.table}" SET expire_time = ? WHERE rowid = ?',
                        (expire_time, rowid),
                    )
                    return True
        return False
                    
    
    async def atouch(self, key: str, expire: Optional[float] = None, retry: bool = False):
        """Touch `key` in cache and update `expire` time.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key for item
        :param float expire: seconds until item expires
            (default None, no expiry)
        :param bool retry: retry if database timeout occurs (default False)
        :return: True if key was touched
        :raises Timeout: if database timeout occurs
        """
        now = time.time()
        expire_time = None if expire is None else now + expire
        async with self._atransact(retry) as sql:
            rows = await (await sql(
                f'SELECT rowid, expire_time FROM "{self.table}"'
                ' WHERE key = ?',
                (key,),
            )).fetchall()
            if rows:
                ((rowid, old_expire_time),) = rows
                if old_expire_time is None or old_expire_time > now:
                    await sql(
                        f'UPDATE "{self.table}" SET expire_time = ? WHERE rowid = ?',
                        (expire_time, rowid),
                    )
                    return True
        return False
                    
    

    def add(self, key: str, value: t.Any, expire: t.Optional[float] = None, read: bool = False, tag: t.Optional[str] = None, retry: bool = False):
        """Add `key` and `value` item to cache.

        Similar to `set`, but only add to cache if key not present.

        Operation is atomic. Only one concurrent add operation for a given key
        will succeed.

        When `read` is `True`, `value` should be a file-like object opened
        for reading in binary mode.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key for item
        :param value: value for item
        :param float expire: seconds until the key expires
            (default None, no expiry)
        :param bool read: read value as bytes from file (default False)
        :param str tag: text to associate with key (default None)
        :param bool retry: retry if database timeout occurs (default False)
        :return: True if item was added
        :raises Timeout: if database timeout occurs

        """
        now = time.time()
        expire_time = None if expire is None else now + expire
        size = self.get_object_size(value)
        # size, mode, filename, db_value = self._disk.store(value, read, key=key)
        # columns = (expire_time, tag, size, mode, filename, db_value)
        columns = (expire_time, tag, size, value)

        # with self._transact(retry, filename) as (sql, cleanup):
        with self._transact(retry) as sql:
            rows = sql(
                f'SELECT rowid, expire_time FROM "{self.table}"'
                ' WHERE key = ?',
                (key,),
            ).fetchall()

            if rows:
                ((rowid, old_expire_time),) = rows
                if old_expire_time is None or old_expire_time > now:
                    return False
                self._row_update(rowid, now, columns)
            else: self._row_insert(key, now, columns)
            self._cull(now, sql)
            return True
    
    async def aadd(self, key: str, value: t.Any, expire: t.Optional[float] = None, read: bool = False, tag: t.Optional[str] = None, retry: bool = False):
        """Add `key` and `value` item to cache.

        Similar to `set`, but only add to cache if key not present.

        Operation is atomic. Only one concurrent add operation for a given key
        will succeed.

        When `read` is `True`, `value` should be a file-like object opened
        for reading in binary mode.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key for item
        :param value: value for item
        :param float expire: seconds until the key expires
            (default None, no expiry)
        :param bool read: read value as bytes from file (default False)
        :param str tag: text to associate with key (default None)
        :param bool retry: retry if database timeout occurs (default False)
        :return: True if item was added
        :raises Timeout: if database timeout occurs

        """
        now = time.time()
        expire_time = None if expire is None else now + expire
        size = self.get_object_size(value)
        columns = (expire_time, tag, size, value)
        async with self._atransact(retry) as sql:
            rows = await (await sql(
                f'SELECT rowid, expire_time FROM "{self.table}"'
                ' WHERE key = ?',
                (key,),
            )).fetchall()

            if rows:
                ((rowid, old_expire_time),) = rows
                if old_expire_time is None or old_expire_time > now:
                    return False
                await self._row_update(rowid, now, columns, is_async = True)
            else: await self._row_insert(key, now, columns, is_async = True)
            self.abg_cull(now, retry = retry)
            # self._cull(now, sql)
            return True
    

    def incr(self, key: str, delta: int = 1, default: int = 0, retry: bool = False):
        """Increment value by delta for item with key.

        If key is missing and default is None then raise KeyError. Else if key
        is missing and default is not None then use default for value.

        Operation is atomic. All concurrent increment operations will be
        counted individually.

        Assumes value may be stored in a SQLite column. Most builds that target
        machines with 64-bit pointer widths will support 64-bit signed
        integers.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key for item
        :param int delta: amount to increment (default 1)
        :param int default: value if key is missing (default 0)
        :param bool retry: retry if database timeout occurs (default False)
        :return: new value for item
        :raises KeyError: if key is not found and default is None
        :raises Timeout: if database timeout occurs

        """
        now = time.time()

        # db_key, raw = self._disk.put(key)
        select = (
            f'SELECT rowid, expire_time, value FROM "{self.table}"'
            ' WHERE key = ?'
        )

        with self._transact(retry) as sql:
            rows = sql(select, (key,)).fetchall()
            if not rows:
                if default is None: raise KeyError(key)
                value = default + delta
                # expire_time, tag, size, value
                columns = (None, None, self.get_object_size(value), value)
                self._row_insert(key, now, columns)
                # columns = (None, None) + self._disk.store(
                #     value, False, key=key
                # )
                self._row_insert(key, now, columns)
                self._cull(now, sql)
                return value
            
            ((rowid, expire_time, value),) = rows
            # ((rowid, expire_time, filename, value),) = rows
            if expire_time is not None and expire_time < now:
                if default is None: raise KeyError(key)

                value = default + delta
                columns = (None, None, self.get_object_size(value), value)
                # columns = (None, None) + self._disk.store(
                #     value, False, key=key
                # )
                self._row_update(rowid, now, columns)
                self._cull(now, sql)
                # cleanup(filename)
                return value

            value += delta
            columns = 'store_time = ?, value = ?'
            update_column = self.eviction_policies['get']
            if update_column is not None:
                columns += f', {update_column.format(now=now)}'
            update = f'UPDATE "{self.table}" SET %s WHERE rowid = ?' % columns
            sql(update, (now, value, rowid))
            return value
    
    async def aincr(self, key: str, delta: int = 1, default: int = 0, retry: bool = False):
        """Increment value by delta for item with key.

        If key is missing and default is None then raise KeyError. Else if key
        is missing and default is not None then use default for value.

        Operation is atomic. All concurrent increment operations will be
        counted individually.

        Assumes value may be stored in a SQLite column. Most builds that target
        machines with 64-bit pointer widths will support 64-bit signed
        integers.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key for item
        :param int delta: amount to increment (default 1)
        :param int default: value if key is missing (default 0)
        :param bool retry: retry if database timeout occurs (default False)
        :return: new value for item
        :raises KeyError: if key is not found and default is None
        :raises Timeout: if database timeout occurs

        """
        now = time.time()
        select = (
            f'SELECT rowid, expire_time, value FROM {self.table}'
            ' WHERE key = ?'
        )

        async with self._atransact(retry) as sql:
            rows = await (await sql(select, (key,))).fetchall()
            if not rows:
                if default is None: raise KeyError(key)
                value = default + delta
                columns = (None, None, self.get_object_size(value), value)
                await self._row_insert(key, now, columns, is_async = True)
                await self._row_insert(key, now, columns, is_async = True)
                self.abg_cull(now, retry = retry)
                return value
            
            ((rowid, expire_time, value),) = rows
            if expire_time is not None and expire_time < now:
                if default is None: raise KeyError(key)
                value = default + delta
                columns = (None, None, self.get_object_size(value), value)
                await self._row_update(rowid, now, columns, is_async = True)
                self.abg_cull(now, retry = retry)
                return value

            value += delta
            columns = 'store_time = ?, value = ?'
            update_column = self.eviction_policies['get']
            if update_column is not None:
                columns += f', {update_column.format(now=now)}'
            update = f'UPDATE "{self.table}" SET %s WHERE rowid = ?' % columns
            await sql(update, (now, value, rowid))
            return value

    def decr(self, key: str, delta: int = 1, default: int = 0, retry: bool = False):
        """Decrement value by delta for item with key.

        If key is missing and default is None then raise KeyError. Else if key
        is missing and default is not None then use default for value.

        Operation is atomic. All concurrent decrement operations will be
        counted individually.

        Unlike Memcached, negative values are supported. Value may be
        decremented below zero.

        Assumes value may be stored in a SQLite column. Most builds that target
        machines with 64-bit pointer widths will support 64-bit signed
        integers.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key for item
        :param int delta: amount to decrement (default 1)
        :param int default: value if key is missing (default 0)
        :param bool retry: retry if database timeout occurs (default False)
        :return: new value for item
        :raises KeyError: if key is not found and default is None
        :raises Timeout: if database timeout occurs

        """
        return self.incr(key, -delta, default, retry)

    
    async def adecr(self, key: str, delta: int = 1, default: int = 0, retry: bool = False):
        """Decrement value by delta for item with key.

        If key is missing and default is None then raise KeyError. Else if key
        is missing and default is not None then use default for value.

        Operation is atomic. All concurrent decrement operations will be
        counted individually.

        Unlike Memcached, negative values are supported. Value may be
        decremented below zero.

        Assumes value may be stored in a SQLite column. Most builds that target
        machines with 64-bit pointer widths will support 64-bit signed
        integers.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key for item
        :param int delta: amount to decrement (default 1)
        :param int default: value if key is missing (default 0)
        :param bool retry: retry if database timeout occurs (default False)
        :return: new value for item
        :raises KeyError: if key is not found and default is None
        :raises Timeout: if database timeout occurs

        """
        return await self.aincr(key, -delta, default, retry)

    def get(
        self,
        key: str,
        default: t.Optional[t.Any] = None,
        include_meta: t.Optional[bool] = False,
        retry: bool = False,
    ) -> t.Union[t.Any, t.Tuple[t.Any, t.Dict[str, t.Any]]]:
        """Retrieve value from cache. If `key` is missing, return `default`.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key for item
        :param default: value to return if key is missing (default None)
        :param include_meta: if True, return a dict of metadata as the second item (default False)
        :param bool retry: retry if database timeout occurs (default False)
        :return: value for item or default if key not found or a tuple of (value, metadata)
        :raises Timeout: if database timeout occurs
        """

        # db_key, raw = self._disk.put(key)
        update_column = self.eviction_policies['get']
        select = (
            'SELECT rowid, expire_time, tag, access_count, hit_count, value'
            f' FROM "{self.table}" WHERE key = ?'
            ' AND (expire_time IS NULL OR expire_time > ?)'
        )
        if not self.statistics and update_column is None:
            # Fast path, no transaction necessary.
            rows = self._sql(select, (key, time.time())).fetchall()
            if not rows: return default

            # ((rowid, db_expire_time, db_tag, mode, filename, db_value),) = rows
            ((rowid, expire_time, tag, access_count, hit_count, value),) = rows

        else:  # Slow path, transaction required.
            _result = self._update_statistics(key, select, update_column, retry = retry)
            if _result is ENOVAL: return default
            ((rowid, expire_time, tag, access_count, hit_count, value),) = _result

        if not include_meta: return value
        return value, {
            'expire_time': expire_time,
            'rowid': rowid,
            'access_count': access_count,
            'hits': hit_count,
            'tag': tag,
        }


    async def aget(
        self,
        key: str,
        default: t.Optional[t.Any] = None,
        include_meta: t.Optional[bool] = False,
        retry: bool = False,
    ) -> t.Union[t.Any, t.Tuple[t.Any, t.Dict[str, t.Any]]]:
        """Retrieve value from cache. If `key` is missing, return `default`.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key for item
        :param default: value to return if key is missing (default None)
        :param include_meta: if True, return a dict of metadata as the second item (default False)
        :param bool retry: retry if database timeout occurs (default False)
        :return: value for item or default if key not found or a tuple of (value, metadata)
        :raises Timeout: if database timeout occurs
        """
        update_column = self.eviction_policies['get']
        select = (
            'SELECT rowid, expire_time, tag, access_count, hit_count, value'
            f' FROM "{self.table}" WHERE key = ?'
            ' AND (expire_time IS NULL OR expire_time > ?)'
        )
        if not self.statistics and update_column is None:
            rows = await (await self._asql(select, (key, time.time()))).fetchall()
            if not rows: return default
            ((rowid, expire_time, tag, access_count, hit_count, value),) = rows

        else:  # Slow path, transaction required.
            _result = await self._aupdate_statistics(key, select, update_column, retry = retry)
            if _result is ENOVAL: return default
            ((rowid, expire_time, tag, access_count, hit_count, value),) = _result
        
        if not include_meta: return value
        return value, {
            'expire_time': expire_time,
            'rowid': rowid,
            'access_count': access_count,
            'hits': hit_count,
            'tag': tag,
        }

    
    def _update_statistics(self, key: str, select: str, update_column: t.Optional[str] = None, retry: bool = False):
        """
        Update The Statics
        """
        cache_hit = (
            f'UPDATE "{self.table}_settings" SET value = value + 1 WHERE key = "hits"'
        )
        cache_miss = (
            f'UPDATE "{self.table}_settings" SET value = value + 1 WHERE key = "misses"'
        )
        with self._transact(retry) as sql:
            rows = sql(select, (key, time.time())).fetchall()
            if not rows:
                if self.statistics: sql(cache_miss)
                return ENOVAL
            (
                (rowid, _, _, _, _, _),
            ) = rows
            if self.statistics: sql(cache_hit)
            now = time.time()
            update = f'UPDATE "{self.table}" SET %s WHERE rowid = ?'
            if update_column is not None:
                sql(update % update_column.format(now=now), (rowid,))
            return rows

    async def _aupdate_statistics(self, key: str, select: str, update_column: t.Optional[str] = None, retry: bool = False):
        """
        [Async Background] Update The Statics
        """
        cache_hit = (
            f'UPDATE "{self.table}_settings" SET value = value + 1 WHERE key = "hits"'
        )
        cache_miss = (
            f'UPDATE "{self.table}_settings" SET value = value + 1 WHERE key = "misses"'
        )
        async with self._atransact(retry) as sql:
            rows = await (await sql(select, (key, time.time()))).fetchall()
            if not rows:
                if self.statistics: await sql(cache_miss)
                return ENOVAL
            (
                (rowid, _, _, _, _, _),
            ) = rows
            if self.statistics: await sql(cache_hit)

            now = time.time()
            update = f'UPDATE "{self.table}" SET %s WHERE rowid = ?'
            if update_column is not None:
                await sql(update % update_column.format(now=now), (rowid,))
            return rows

    def fetch_kv_data(self, limit: t.Optional[int] = None, **kwargs) -> t.Dict[str, t.Any]:
        """
        Loads all the Data
        """
        # Since this is a heavy operation, we'll just retrieve it from the DB
        # with a single query
        select = (
            'SELECT key, value'
            f' FROM "{self.table}"'
            ' WHERE (expire_time IS NULL OR expire_time > ?)'
        )
        if limit is not None: select += f' LIMIT {limit}'
        now = time.time()
        data_results = {}
        rows = self._sql(select, (now,)).fetchall()
        if not rows: return data_results
        for row in rows:
            key, value = row
            data_results[key] = value
        return data_results

    async def afetch_kv_data(self, limit: t.Optional[int] = None, **kwargs) -> t.Dict[str, t.Any]:
        """
        Loads all the Data
        """
        # Since this is a heavy operation, we'll just retrieve it from the DB
        # with a single query
        select = (
            'SELECT key, value'
            f' FROM "{self.table}"'
            ' WHERE (expire_time IS NULL OR expire_time > ?)'
        )
        if limit is not None: select += f' LIMIT {limit}'
        now = time.time()
        data_results = {}
        async with self._atransact() as sql:
            rows = await (await sql(select, (now,))).fetchall()
            if not rows: return data_results
            for row in rows:
                key, value = row
                data_results[key] = value
        return data_results


    def fetch_values(self, limit: t.Optional[int] = None, **kwargs) -> t.List[t.Any]:
        """
        Fetches the values
        """
        select = (
            'SELECT value'
            f' FROM "{self.table}"'
            ' WHERE (expire_time IS NULL OR expire_time > ?)'
        )
        if limit is not None: select += f' LIMIT {limit}'
        now = time.time()
        rows = self._sql(select, (now,)).fetchall()
        return [row[0] for row in rows]

    async def afetch_values(self, limit: t.Optional[int] = None, **kwargs) -> t.List[t.Any]:
        """
        Fetches the values
        """
        select = (
            'SELECT value'
            f' FROM "{self.table}"'
            ' WHERE (expire_time IS NULL OR expire_time > ?)'
        )
        if limit is not None: select += f' LIMIT {limit}'
        now = time.time()
        rows = await (await self._asql(select, (now,))).fetchall()
        return [row[0] for row in rows]
        
    def batch_get(self, *keys: str, default: t.Optional[t.Any] = None, include_meta: bool = False, retry: t.Optional[bool] = False):
        """
        Batch gets many items

        :param keys: keys to get
        :param default: value to return if key is missing
            (default None, no default value)
        :param include_meta: if True, return a dict of metadata
            (default False)
        :param retry: retry if database timeout occurs (default None, no retry)
        :return: dict of key/value pairs or tuple of (key, value) pairs if include_meta is True
        """
        update_column = self.eviction_policies['get']
        select = (
            'SELECT rowid, expire_time, tag, access_count, hit_count, value, key'
            f' FROM "{self.table}" WHERE key IN (%s)'
            ' AND (expire_time IS NULL OR expire_time > ?)'
        ) % self._format_keys_(keys)
        now = time.time()
        meta_results = {}
        batch_results = {}
        def return_noval():
            _batch_results = {key: default for key in keys}
            if include_meta: 
                _meta_results = {key: None for key in keys}
                return _batch_results, _meta_results
            return _batch_results
            
        if not self.statistics and update_column is None:
            # Fast path, no transaction necessary.
            rows = self._sql(select, (now,)).fetchall()
            if not rows:  return return_noval()
            fetched_results = {
                key: (rowid, value, db_expire_time, db_tag, db_access_count, db_hit_count)
                for (rowid, db_expire_time, db_tag, db_access_count, db_hit_count, value, key) in rows
            }
            for key in keys:
                if key not in fetched_results:
                    batch_results[key] = default
                    if include_meta: meta_results[key] = None
                    continue
                (rowid, value, db_expire_time, db_tag, db_access_count, db_hit_count) = fetched_results[key]
                if include_meta: meta_results[key] = {
                    'expire_time': db_expire_time,
                    'tag': db_tag,
                    'rowid': rowid,
                    'access_count': db_access_count,
                    'hits': db_hit_count,
                }
                batch_results[key] = value

        else:  # Slow path, transaction required.
            _result = self._batch_update_statistics(keys, now, select, update_column = update_column, default = default, retry = retry)
            if _result is ENOVAL: return return_noval()
            batch_results, meta_results = _result

        return (batch_results, meta_results) if include_meta else batch_results


    async def abatch_get(self, *keys: str, default: t.Optional[t.Any] = None, include_meta: bool = False, retry: t.Optional[bool] = False):
        """
        Batch gets many items

        :param keys: keys to get
        :param default: value to return if key is missing
            (default None, no default value)
        :param include_meta: if True, return a dict of metadata
            (default False)
        :param retry: retry if database timeout occurs (default None, no retry)
        :return: dict of key/value pairs or tuple of (key, value) pairs if include_meta is True
        """
        update_column = self.eviction_policies['get']
        select = (
            'SELECT rowid, expire_time, tag, access_count, hit_count, value, key'
            f' FROM "{self.table}" WHERE key IN (%s)'
            ' AND (expire_time IS NULL OR expire_time > ?)'
        ) % self._format_keys_(keys)
        now = time.time()
        meta_results = {}
        batch_results = {}
        def return_noval():
            _batch_results = {key: default for key in keys}
            if include_meta: 
                _meta_results = {key: None for key in keys}
                return _batch_results, _meta_results
            return _batch_results
            
        if not self.statistics and update_column is None:
            # Fast path, no transaction necessary.
            rows = await (await self._asql(select, (now,))).fetchall()
            if not rows:  return return_noval()
            fetched_results = {
                key: (rowid, value, db_expire_time, db_tag, db_access_count, db_hit_count)
                for (rowid, db_expire_time, db_tag, db_access_count, db_hit_count, value, key) in rows
            }
            for key in keys:
                if key not in fetched_results:
                    batch_results[key] = default
                    if include_meta: meta_results[key] = None
                    continue
                (rowid, value, db_expire_time, db_tag, db_access_count, db_hit_count) = fetched_results[key]
                if include_meta: meta_results[key] = {
                    'expire_time': db_expire_time,
                    'tag': db_tag,
                    'rowid': rowid,
                    'access_count': db_access_count,
                    'hits': db_hit_count,
                }
                batch_results[key] = value

        else:  # Slow path, transaction required.
            _result = await self._abatch_update_statistics(keys, now, select, update_column = update_column, default = default, retry = retry)
            if _result is ENOVAL: return return_noval()
            batch_results, meta_results = _result

        return (batch_results, meta_results) if include_meta else batch_results

    def _batch_update_statistics(self, keys: t.Iterable[str], now: float, select: str, update_column: t.Optional[str] = None, default: t.Optional[t.Any] = None, retry: bool = False):
        """
        Batch Update The Statics and get the values
        """
        cache_hit = (
            f'UPDATE "{self.table}_settings" SET value = value + <NUM_HITS> WHERE key = "hits"'
        )
        cache_miss = (
            f'UPDATE "{self.table}_settings" SET value = value + <NUM_MISSES> WHERE key = "misses"'
        )
        cache_bulk_miss = (
            f'UPDATE "{self.table}_settings" SET value = value + {len(keys)} WHERE key = "misses"'
        )
        with self._transact(retry) as sql:
            rows = sql(select, (now,)).fetchall()
            if not rows:
                if self.statistics: sql(cache_bulk_miss)
                return ENOVAL
            fetched_results = {
                key: (rowid, value, db_expire_time, db_tag, db_access_count, db_hit_count)
                for (rowid, db_expire_time, db_tag, db_access_count, db_hit_count, value, key) in rows
            }
            meta_results, batch_results = {}, {}
            n_hits, n_misses = 0, 0
            row_ids = []
            for key in keys:
                if key not in fetched_results:
                    batch_results[key] = default
                    meta_results[key] = None
                    n_misses += 1
                    continue
                (rowid, value, db_expire_time, db_tag, db_access_count, db_hit_count) = fetched_results[key]
                meta_results[key] = {
                    'expire_time': db_expire_time,
                    'tag': db_tag,
                    'rowid': rowid,
                    'access_count': db_access_count,
                    'hits': db_hit_count,
                }
                batch_results[key] = value
                n_hits += 1
                # row_ids.append((rowid,))
                row_ids.append(rowid)

            if self.statistics: 
                if n_hits: sql(cache_hit.replace('<NUM_HITS>', str(n_hits)))
                if n_misses: sql(cache_miss.replace('<NUM_MISSES>', str(n_misses)))
            _row_ids = self._format_rowids_(row_ids)
            update = f'UPDATE "{self.table}" SET %s WHERE rowid IN ({_row_ids})'
            if update_column is not None:
                sql(update % update_column.format(now=now))
                # sqlmany(update % update_column.format(now=now), row_ids)
            return batch_results, meta_results

    async def _abatch_update_statistics(self, keys: t.Iterable[str], now: float, select: str, update_column: t.Optional[str] = None, default: t.Optional[t.Any] = None, retry: bool = False):
        """
        [Async] Batch Update The Statics and get the values
        """
        cache_hit = (
            f'UPDATE "{self.table}_settings" SET value = value + <NUM_HITS> WHERE key = "hits"'
        )
        cache_miss = (
            f'UPDATE "{self.table}_settings" SET value = value + <NUM_MISSES> WHERE key = "misses"'
        )
        cache_bulk_miss = (
            f'UPDATE "{self.table}_settings" SET value = value + {len(keys)} WHERE key = "misses"'
        )
        async with self._atransact(retry) as sql:
            rows = await (await sql(select, (now,))).fetchall()
            if not rows:
                if self.statistics: await sql(cache_bulk_miss)
                return ENOVAL
            fetched_results = {
                key: (rowid, value, db_expire_time, db_tag, db_access_count, db_hit_count)
                for (rowid, db_expire_time, db_tag, db_access_count, db_hit_count, value, key) in rows
            }
            meta_results, batch_results = {}, {}
            n_hits, n_misses = 0, 0
            row_ids = []
            for key in keys:
                if key not in fetched_results:
                    batch_results[key] = default
                    meta_results[key] = None
                    n_misses += 1
                    continue
                (rowid, value, db_expire_time, db_tag, db_access_count, db_hit_count) = fetched_results[key]
                meta_results[key] = {
                    'expire_time': db_expire_time,
                    'tag': db_tag,
                    'rowid': rowid,
                    'access_count': db_access_count,
                    'hits': db_hit_count,
                }
                batch_results[key] = value
                n_hits += 1
                row_ids.append(rowid)

            if self.statistics: 
                if n_hits: await sql(cache_hit.replace('<NUM_HITS>', str(n_hits)))
                if n_misses: await sql(cache_miss.replace('<NUM_MISSES>', str(n_misses)))
            _row_ids = self._format_rowids_(row_ids)
            update = f'UPDATE "{self.table}" SET %s WHERE rowid IN ({_row_ids})'
            if update_column is not None:
                await sql(update % update_column.format(now=now))
            return batch_results, meta_results


    def select_tags(
        self, 
        *tags: str, 
        limit: t.Optional[int] = None, 
        order: t.Optional[t.Literal['ASC', 'DESC']] = 'ASC',
        order_by: t.Optional[str] = 'rowid',
        include_meta: bool = False, 
        retry: t.Optional[bool] = True
    ) -> t.Union[t.Dict[str, t.Any], t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]]:
        """
        Select all items with the given tags

        :param tags: tags to select
        :param limit: limit the number of items to return
            (default None, no limit)
        :param order: order to return the items
            (default 'ASC', ascending)
        :param include_meta: if True, return a dict of metadata
            (default False)
        :param retry: retry if database timeout occurs (default True)
        :return: dict of key/value pairs or tuple of (key, value) pairs if include_meta is True
        """
        update_column = self.eviction_policies['get']
        select = (
            f'SELECT rowid, expire_time, tag, access_count, hit_count, value, key'
            f' FROM "{self.table}"'
            ' WHERE tag IN (%s) AND (expire_time IS NULL OR expire_time > ?)'
        ) % self._format_keys_(tags)
        
        if order: select += f' ORDER BY {order_by} {order}'
        if limit: select += f' LIMIT {limit}'
        now = time.time()
        meta_results, batch_results = {}, {}
        rowids = []
        with self._transact(retry) as sql:
            rows = sql(select, (now,)).fetchall()
            if not rows: return {}, {} if include_meta else {}
            for row in rows:
                rowid, expire_time, tag, access_count, hit_count, value, key = row
                batch_results[key] = value
                if include_meta: meta_results[key] = {
                    'expire_time': expire_time,
                    'tag': tag,
                    'rowid': rowid,
                    'access_count': access_count,
                    'hits': hit_count,
                }
                rowids.append(rowid)
            if self.statistics:
                cache_hit = (
                    f'UPDATE "{self.table}_settings" SET value = value + {len(rowids)} WHERE key = "hits"'
                )
                sql(cache_hit)
            
            if update_column is not None:
                _row_ids = self._format_rowids_(rowids)
                update = f'UPDATE "{self.table}" SET %s WHERE rowid IN ({_row_ids})'
                sql(update % update_column.format(now=now))
        return (batch_results, meta_results) if include_meta else batch_results

    
    async def aselect_tags(
        self, 
        *tags: str, 
        limit: t.Optional[int] = None, 
        order: t.Optional[t.Literal['ASC', 'DESC']] = 'ASC',
        order_by: t.Optional[str] = 'rowid',
        include_meta: bool = False, 
        retry: t.Optional[bool] = True
    ) -> t.Union[t.Dict[str, t.Any], t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]]:
        """
        Select all items with the given tags

        :param tags: tags to select
        :param limit: limit the number of items to return
            (default None, no limit)
        :param order: order to return the items
            (default 'ASC', ascending)
        :param include_meta: if True, return a dict of metadata
            (default False)
        :param retry: retry if database timeout occurs (default True)
        :return: dict of key/value pairs or tuple of (key, value) pairs if include_meta is True
        """
        update_column = self.eviction_policies['get']
        select = (
            f'SELECT rowid, expire_time, tag, access_count, hit_count, value, key'
            f' FROM "{self.table}"'
            ' WHERE tag IN (%s) AND (expire_time IS NULL OR expire_time > ?)'
        ) % self._format_keys_(tags)
        
        if order: select += f' ORDER BY {order_by} {order}'
        if limit: select += f' LIMIT {limit}'
        now = time.time()
        meta_results, batch_results = {}, {}
        rowids = []
        async with self._atransact(retry) as sql:
            rows = await (await sql(select, (now,))).fetchall()
            if not rows: return {}, {} if include_meta else {}
            for row in rows:
                rowid, expire_time, tag, access_count, hit_count, value, key = row
                batch_results[key] = value
                if include_meta: meta_results[key] = {
                    'expire_time': expire_time,
                    'tag': tag,
                    'rowid': rowid,
                    'access_count': access_count,
                    'hits': hit_count,
                }
                rowids.append(rowid)
            if self.statistics:
                cache_hit = (
                    f'UPDATE "{self.table}_settings" SET value = value + {len(rowids)} WHERE key = "hits"'
                )
                await sql(cache_hit)
            
            if update_column is not None:
                _row_ids = self._format_rowids_(rowids)
                update = f'UPDATE "{self.table}" SET %s WHERE rowid IN ({_row_ids})'
                await sql(update % update_column.format(now=now))
        return (batch_results, meta_results) if include_meta else batch_results

        

    def read(self, key: str, retry: bool = False):
        """Return file handle value corresponding to `key` from cache.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key matching item
        :param bool retry: retry if database timeout occurs (default False)
        :return: file open for reading in binary mode
        :raises KeyError: if key is not found
        :raises Timeout: if database timeout occurs

        """
        handle = self.get(key, default=ENOVAL, read=True, retry=retry)
        if handle is ENOVAL:
            raise KeyError(key)
        return handle


    def pop(
        self, key: str, default: t.Optional[t.Any] = None, expire_time: bool = False, tag: bool = False, retry: bool = False
    ):  # noqa: E501
        """Remove corresponding item for `key` from cache and return value.

        If `key` is missing, return `default`.

        Operation is atomic. Concurrent operations will be serialized.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key for item
        :param default: value to return if key is missing (default None)
        :param bool expire_time: if True, return expire_time in tuple
            (default False)
        :param bool tag: if True, return tag in tuple (default False)
        :param bool retry: retry if database timeout occurs (default False)
        :return: value for item or default if key not found
        :raises Timeout: if database timeout occurs

        """
        # db_key, raw = self._disk.put(key)
        select = (
            'SELECT rowid, expire_time, tag, value'
            f' FROM "{self.table}" WHERE key = ?'
            ' AND (expire_time IS NULL OR expire_time > ?)'
        )

        if expire_time and tag: default = default, None, None
        elif expire_time or tag: default = default, None
        with self._transact(retry) as sql:
            rows = sql(select, (key, time.time())).fetchall()
            # rows = sql(select, (db_key, raw, time.time())).fetchall()
            if not rows: return default
            ((rowid, db_expire_time, db_tag, value),) = rows
            # ((rowid, db_expire_time, db_tag, mode, filename, db_value),) = rows
            sql(f'DELETE FROM "{self.table}" WHERE rowid = ?', (rowid,))

        if expire_time and tag: return value, db_expire_time, db_tag
        elif expire_time: return value, db_expire_time
        elif tag: return value, db_tag
        else: return value
    
    
    async def apop(
        self, key: str, default: t.Optional[t.Any] = None, expire_time: bool = False, tag: bool = False, retry: bool = False
    ):  # noqa: E501
        """Remove corresponding item for `key` from cache and return value.

        If `key` is missing, return `default`.

        Operation is atomic. Concurrent operations will be serialized.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key for item
        :param default: value to return if key is missing (default None)
        :param bool expire_time: if True, return expire_time in tuple
            (default False)
        :param bool tag: if True, return tag in tuple (default False)
        :param bool retry: retry if database timeout occurs (default False)
        :return: value for item or default if key not found
        :raises Timeout: if database timeout occurs

        """
        select = (
            'SELECT rowid, expire_time, tag, value'
            f' FROM "{self.table}" WHERE key = ?'
            ' AND (expire_time IS NULL OR expire_time > ?)'
        )

        if expire_time and tag: default = default, None, None
        elif expire_time or tag: default = default, None
        async with self._atransact(retry) as sql:
            rows = await (await sql(select, (key, time.time()))).fetchall()
            if not rows: return default
            ((rowid, db_expire_time, db_tag, value),) = rows
            await sql(f'DELETE FROM "{self.table}" WHERE rowid = ?', (rowid,))

        if expire_time and tag: return value, db_expire_time, db_tag
        elif expire_time: return value, db_expire_time
        elif tag: return value, db_tag
        else: return value


    def delete(self, key: str, retry: bool = False):
        """Delete corresponding item for `key` from cache.

        Missing keys are ignored.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key matching item
        :param bool retry: retry if database timeout occurs (default False)
        :return: True if item was deleted
        :raises Timeout: if database timeout occurs

        """
        # pylint: disable=unnecessary-dunder-call
        try:
            return self._delete_key_(key, retry=retry)
        except KeyError:
            return False
    
    async def adelete(self, key: str, retry: bool = False):
        """Delete corresponding item for `key` from cache.

        Missing keys are ignored.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param key: key matching item
        :param bool retry: retry if database timeout occurs (default False)
        :return: True if item was deleted
        :raises Timeout: if database timeout occurs

        """
        # pylint: disable=unnecessary-dunder-call
        try:
            return await self._adelete_key_(key, retry=retry)
        except KeyError:
            return False
    
    def push(
        self,
        value: t.Any,
        prefix: t.Optional[str] = None,
        side: t.Literal['back', 'front'] = 'back',
        expire: t.Optional[float] = None,
        read: bool = False,
        tag: t.Optional[str] = None,
        retry: bool = False,
    ):
        """Push `value` onto `side` of queue identified by `prefix` in cache.

        When prefix is None, integer keys are used. Otherwise, string keys are
        used in the format "prefix-integer". Integer starts at 500 trillion.

        Defaults to pushing value on back of queue. Set side to 'front' to push
        value on front of queue. Side must be one of 'back' or 'front'.

        Operation is atomic. Concurrent operations will be serialized.

        When `read` is `True`, `value` should be a file-like object opened
        for reading in binary mode.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        See also `Cache.pull`.

        >>> cache = Cache()
        >>> print(cache.push('first value'))
        500000000000000
        >>> cache.get(500000000000000)
        'first value'
        >>> print(cache.push('second value'))
        500000000000001
        >>> print(cache.push('third value', side='front'))
        499999999999999
        >>> cache.push(1234, prefix='userids')
        'userids-500000000000000'

        :param value: value for item
        :param str prefix: key prefix (default None, key is integer)
        :param str side: either 'back' or 'front' (default 'back')
        :param float expire: seconds until the key expires
            (default None, no expiry)
        :param bool read: read value as bytes from file (default False)
        :param str tag: text to associate with key (default None)
        :param bool retry: retry if database timeout occurs (default False)
        :return: key for item in cache
        :raises Timeout: if database timeout occurs

        """
        if prefix is None:
            min_key = 0
            max_key = 999999999999999
        else:
            min_key = f'{prefix}-000000000000000'
            max_key = f'{prefix}-999999999999999'

        now = time.time()
        # raw = True
        expire_time = None if expire is None else now + expire
        size = self.get_object_size(value)
        columns = (expire_time, tag, size, value)
        # size, mode, filename, db_value = self._disk.store(value, read)
        # columns = (expire_time, tag, size, mode, filename, db_value)
        order = {'back': 'DESC', 'front': 'ASC'}
        select = (
            f'SELECT key FROM "{self.table}"'
            ' WHERE ? < key AND key < ?'
            ' ORDER BY key %s LIMIT 1'
        ) % order[side]
        with self._transact(retry) as sql:
            rows = sql(select, (min_key, max_key)).fetchall()
            if rows:
                ((key,),) = rows
                key: str

                num = int(key[(key.rfind('-') + 1) :]) if prefix is not None else key
                if side == 'back': num += 1
                else:
                    assert side == 'front'
                    num -= 1
            else:
                num = 500000000000000
            db_key = '{0}-{1:015d}'.format(prefix, num) if prefix is not None else num
            self._row_insert(db_key, now, columns)
            self._cull(now, sql)
            return db_key
    
    async def apush(
        self,
        value: t.Any,
        prefix: t.Optional[str] = None,
        side: t.Literal['back', 'front'] = 'back',
        expire: t.Optional[float] = None,
        read: bool = False,
        tag: t.Optional[str] = None,
        retry: bool = False,
    ):
        """Push `value` onto `side` of queue identified by `prefix` in cache.

        When prefix is None, integer keys are used. Otherwise, string keys are
        used in the format "prefix-integer". Integer starts at 500 trillion.

        Defaults to pushing value on back of queue. Set side to 'front' to push
        value on front of queue. Side must be one of 'back' or 'front'.

        Operation is atomic. Concurrent operations will be serialized.

        When `read` is `True`, `value` should be a file-like object opened
        for reading in binary mode.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        See also `Cache.pull`.

        >>> cache = Cache()
        >>> print(cache.push('first value'))
        500000000000000
        >>> cache.get(500000000000000)
        'first value'
        >>> print(cache.push('second value'))
        500000000000001
        >>> print(cache.push('third value', side='front'))
        499999999999999
        >>> cache.push(1234, prefix='userids')
        'userids-500000000000000'

        :param value: value for item
        :param str prefix: key prefix (default None, key is integer)
        :param str side: either 'back' or 'front' (default 'back')
        :param float expire: seconds until the key expires
            (default None, no expiry)
        :param bool read: read value as bytes from file (default False)
        :param str tag: text to associate with key (default None)
        :param bool retry: retry if database timeout occurs (default False)
        :return: key for item in cache
        :raises Timeout: if database timeout occurs

        """
        if prefix is None:
            min_key = 0
            max_key = 999999999999999
        else:
            min_key = f'{prefix}-000000000000000'
            max_key = f'{prefix}-999999999999999'

        now = time.time()
        expire_time = None if expire is None else now + expire
        size = self.get_object_size(value)
        columns = (expire_time, tag, size, value)
        order = {'back': 'DESC', 'front': 'ASC'}
        select = (
            f'SELECT key FROM "{self.table}"'
            ' WHERE ? < key AND key < ?'
            ' ORDER BY key %s LIMIT 1'
        ) % order[side]
        async with self._atransact(retry) as sql:
            rows = await (await sql(select, (min_key, max_key))).fetchall()
            if rows:
                ((key,),) = rows
                key: str
                num = int(key[(key.rfind('-') + 1) :]) if prefix is not None else key
                if side == 'back': num += 1
                else:
                    assert side == 'front'
                    num -= 1
            else:
                num = 500000000000000
            db_key = '{0}-{1:015d}'.format(prefix, num) if prefix is not None else num
            await self._row_insert(db_key, now, columns, is_async = True)
            # self._cull(now, sql)
            self.abg_cull(now, retry = retry)
            return db_key

    def pull(
        self,
        prefix: t.Optional[str] = None,
        default: t.Tuple[t.Any, t.Any] = (None, None),
        side: t.Literal['front', 'back'] = 'front',
        expire_time: bool = False,
        tag: bool = False,
        retry: bool = False,
    ):
        """Pull key and value item pair from `side` of queue in cache.

        When prefix is None, integer keys are used. Otherwise, string keys are
        used in the format "prefix-integer". Integer starts at 500 trillion.

        If queue is empty, return default.

        Defaults to pulling key and value item pairs from front of queue. Set
        side to 'back' to pull from back of queue. Side must be one of 'front'
        or 'back'.

        Operation is atomic. Concurrent operations will be serialized.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        See also `Cache.push` and `Cache.get`.

        >>> cache = Cache()
        >>> cache.pull()
        (None, None)
        >>> for letter in 'abc':
        ...     print(cache.push(letter))
        500000000000000
        500000000000001
        500000000000002
        >>> key, value = cache.pull()
        >>> print(key)
        500000000000000
        >>> value
        'a'
        >>> _, value = cache.pull(side='back')
        >>> value
        'c'
        >>> cache.push(1234, 'userids')
        'userids-500000000000000'
        >>> _, value = cache.pull('userids')
        >>> value
        1234

        :param str prefix: key prefix (default None, key is integer)
        :param default: value to return if key is missing
            (default (None, None))
        :param str side: either 'front' or 'back' (default 'front')
        :param bool expire_time: if True, return expire_time in tuple
            (default False)
        :param bool tag: if True, return tag in tuple (default False)
        :param bool retry: retry if database timeout occurs (default False)
        :return: key and value item pair or default if queue is empty
        :raises Timeout: if database timeout occurs

        """
        # Caution: Nearly identical code exists in Cache.peek
        if prefix is None:
            min_key = 0
            max_key = 999999999999999
        else:
            min_key = f'{prefix}-000000000000000'
            max_key = f'{prefix}-999999999999999'

        order = {'front': 'ASC', 'back': 'DESC'}
        select = (
            'SELECT rowid, key, expire_time, tag, value'
            f' FROM "{self.table}" WHERE ? < key AND key < ?'
            ' ORDER BY key %s LIMIT 1'
        ) % order[side]

        if expire_time and tag: default = default, None, None
        elif expire_time or tag: default = default, None
        while True:
            while True:
                # with self._transact(retry) as (sql, cleanup):
                with self._transact(retry) as sql:
                    rows = sql(select, (min_key, max_key)).fetchall()
                    if not rows: return default
                    (
                        (rowid, key, db_expire, db_tag, value),
                    ) = rows
                    # (
                    #     (rowid, key, db_expire, db_tag, mode, name, db_value),
                    # ) = rows

                    sql(f'DELETE FROM "{self.table}" WHERE rowid = ?', (rowid,))
                    break
                    # if db_expire is not None and db_expire < time.time():
                    #     cleanup(name)
                    # else:
                    #     break

            # try:
            #     value = self._disk.fetch(mode, name, db_value, False)
            # except IOError:
            #     # Key was deleted before we could retrieve result.
            #     continue
            # finally:
            #     if name is not None:
            #         self._disk.remove(name)
            break

        if expire_time and tag: return (key, value), db_expire, db_tag
        elif expire_time: return (key, value), db_expire
        elif tag: return (key, value), db_tag
        return key, value
    

    async def apull(
        self,
        prefix: t.Optional[str] = None,
        default: t.Tuple[t.Any, t.Any] = (None, None),
        side: t.Literal['front', 'back'] = 'front',
        expire_time: bool = False,
        tag: bool = False,
        retry: bool = False,
    ):
        """Pull key and value item pair from `side` of queue in cache.

        When prefix is None, integer keys are used. Otherwise, string keys are
        used in the format "prefix-integer". Integer starts at 500 trillion.

        If queue is empty, return default.

        Defaults to pulling key and value item pairs from front of queue. Set
        side to 'back' to pull from back of queue. Side must be one of 'front'
        or 'back'.

        Operation is atomic. Concurrent operations will be serialized.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        See also `Cache.push` and `Cache.get`.

        >>> cache = Cache()
        >>> cache.pull()
        (None, None)
        >>> for letter in 'abc':
        ...     print(cache.push(letter))
        500000000000000
        500000000000001
        500000000000002
        >>> key, value = cache.pull()
        >>> print(key)
        500000000000000
        >>> value
        'a'
        >>> _, value = cache.pull(side='back')
        >>> value
        'c'
        >>> cache.push(1234, 'userids')
        'userids-500000000000000'
        >>> _, value = cache.pull('userids')
        >>> value
        1234

        :param str prefix: key prefix (default None, key is integer)
        :param default: value to return if key is missing
            (default (None, None))
        :param str side: either 'front' or 'back' (default 'front')
        :param bool expire_time: if True, return expire_time in tuple
            (default False)
        :param bool tag: if True, return tag in tuple (default False)
        :param bool retry: retry if database timeout occurs (default False)
        :return: key and value item pair or default if queue is empty
        :raises Timeout: if database timeout occurs

        """
        # Caution: Nearly identical code exists in Cache.peek
        if prefix is None:
            min_key = 0
            max_key = 999999999999999
        else:
            min_key = f'{prefix}-000000000000000'
            max_key = f'{prefix}-999999999999999'

        order = {'front': 'ASC', 'back': 'DESC'}
        select = (
            'SELECT rowid, key, expire_time, tag, value'
            f' FROM "{self.table}" WHERE ? < key AND key < ?'
            ' ORDER BY key %s LIMIT 1'
        ) % order[side]

        if expire_time and tag: default = default, None, None
        elif expire_time or tag: default = default, None
        while True:
            while True:
                async with self._atransact(retry) as sql:
                    rows = await (await sql(select, (min_key, max_key))).fetchall()
                    if not rows: return default
                    (
                        (rowid, key, db_expire, db_tag, value),
                    ) = rows
                    await sql(f'DELETE FROM "{self.table}" WHERE rowid = ?', (rowid,))
                    break
            break

        if expire_time and tag: return (key, value), db_expire, db_tag
        elif expire_time: return (key, value), db_expire
        elif tag: return (key, value), db_tag
        return key, value

    def peek(
        self,
        prefix: t.Optional[str] = None,
        default: t.Tuple[t.Any, t.Any] = (None, None),
        side: t.Literal['front', 'back'] = 'front',
        expire_time: bool = False,
        tag: bool = False,
        retry: bool = False,
    ):
        """Peek at key and value item pair from `side` of queue in cache.

        When prefix is None, integer keys are used. Otherwise, string keys are
        used in the format "prefix-integer". Integer starts at 500 trillion.

        If queue is empty, return default.

        Defaults to peeking at key and value item pairs from front of queue.
        Set side to 'back' to pull from back of queue. Side must be one of
        'front' or 'back'.

        Expired items are deleted from cache. Operation is atomic. Concurrent
        operations will be serialized.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        See also `Cache.pull` and `Cache.push`.

        >>> cache = Cache()
        >>> for letter in 'abc':
        ...     print(cache.push(letter))
        500000000000000
        500000000000001
        500000000000002
        >>> key, value = cache.peek()
        >>> print(key)
        500000000000000
        >>> value
        'a'
        >>> key, value = cache.peek(side='back')
        >>> print(key)
        500000000000002
        >>> value
        'c'

        :param str prefix: key prefix (default None, key is integer)
        :param default: value to return if key is missing
            (default (None, None))
        :param str side: either 'front' or 'back' (default 'front')
        :param bool expire_time: if True, return expire_time in tuple
            (default False)
        :param bool tag: if True, return tag in tuple (default False)
        :param bool retry: retry if database timeout occurs (default False)
        :return: key and value item pair or default if queue is empty
        :raises Timeout: if database timeout occurs

        """
        # Caution: Nearly identical code exists in Cache.pull
        if prefix is None:
            min_key = 0
            max_key = 999999999999999
        else:
            min_key = f'{prefix}-000000000000000'
            max_key = f'{prefix}-999999999999999'

        order = {'front': 'ASC', 'back': 'DESC'}
        select = (
            'SELECT rowid, key, expire_time, tag, value'
            f' FROM "{self.table}" WHERE ? < key AND key < ?'
            ' ORDER BY key %s LIMIT 1'
        ) % order[side]

        if expire_time and tag: default = default, None, None
        elif expire_time or tag: default = default, None

        while True:
            while True:
                with self._transact(retry) as sql:
                    rows = sql(select, (min_key, max_key)).fetchall()
                    if not rows: return default
                    (
                        (rowid, key, db_expire, db_tag, value),
                    ) = rows
                    # (
                    #     (rowid, key, db_expire, db_tag, mode, name, db_value),
                    # ) = rows

                    if db_expire is None or db_expire >= time.time(): break
                    sql(f'DELETE FROM "{self.table}" WHERE rowid = ?', (rowid,))
                    # cleanup(name)
            # try:
            #     value = self._disk.fetch(mode, name, db_value, False)
            # except IOError:
            #     # Key was deleted before we could retrieve result.
            #     continue
            break

        if expire_time and tag: return (key, value), db_expire, db_tag
        elif expire_time: return (key, value), db_expire
        elif tag: return (key, value), db_tag
        return key, value
    
    async def apeek(
        self,
        prefix: t.Optional[str] = None,
        default: t.Tuple[t.Any, t.Any] = (None, None),
        side: t.Literal['front', 'back'] = 'front',
        expire_time: bool = False,
        tag: bool = False,
        retry: bool = False,
    ):
        """Peek at key and value item pair from `side` of queue in cache.

        When prefix is None, integer keys are used. Otherwise, string keys are
        used in the format "prefix-integer". Integer starts at 500 trillion.

        If queue is empty, return default.

        Defaults to peeking at key and value item pairs from front of queue.
        Set side to 'back' to pull from back of queue. Side must be one of
        'front' or 'back'.

        Expired items are deleted from cache. Operation is atomic. Concurrent
        operations will be serialized.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        See also `Cache.pull` and `Cache.push`.

        >>> cache = Cache()
        >>> for letter in 'abc':
        ...     print(cache.push(letter))
        500000000000000
        500000000000001
        500000000000002
        >>> key, value = cache.peek()
        >>> print(key)
        500000000000000
        >>> value
        'a'
        >>> key, value = cache.peek(side='back')
        >>> print(key)
        500000000000002
        >>> value
        'c'

        :param str prefix: key prefix (default None, key is integer)
        :param default: value to return if key is missing
            (default (None, None))
        :param str side: either 'front' or 'back' (default 'front')
        :param bool expire_time: if True, return expire_time in tuple
            (default False)
        :param bool tag: if True, return tag in tuple (default False)
        :param bool retry: retry if database timeout occurs (default False)
        :return: key and value item pair or default if queue is empty
        :raises Timeout: if database timeout occurs

        """
        # Caution: Nearly identical code exists in Cache.pull
        if prefix is None:
            min_key = 0
            max_key = 999999999999999
        else:
            min_key = f'{prefix}-000000000000000'
            max_key = f'{prefix}-999999999999999'

        order = {'front': 'ASC', 'back': 'DESC'}
        select = (
            'SELECT rowid, key, expire_time, tag, value'
            f' FROM "{self.table}" WHERE ? < key AND key < ?'
            ' ORDER BY key %s LIMIT 1'
        ) % order[side]

        if expire_time and tag: default = default, None, None
        elif expire_time or tag: default = default, None

        while True:
            while True:
                async with self._atransact(retry) as sql:
                    rows = await (await sql(select, (min_key, max_key))).fetchall()
                    if not rows: return default
                    (
                        (rowid, key, db_expire, db_tag, value),
                    ) = rows
                    if db_expire is None or db_expire >= time.time(): break
                    await sql(f'DELETE FROM "{self.table}" WHERE rowid = ?', (rowid,))
            break
        if expire_time and tag: return (key, value), db_expire, db_tag
        elif expire_time: return (key, value), db_expire
        elif tag: return (key, value), db_tag
        return key, value

    def peekitem(self, last: bool = True, expire_time: bool = False, tag: bool = False, retry: bool = False):
        """Peek at key and value item pair in cache based on iteration order.

        Expired items are deleted from cache. Operation is atomic. Concurrent
        operations will be serialized.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        >>> cache = Cache()
        >>> for num, letter in enumerate('abc'):
        ...     cache[letter] = num
        >>> cache.peekitem()
        ('c', 2)
        >>> cache.peekitem(last=False)
        ('a', 0)

        :param bool last: last item in iteration order (default True)
        :param bool expire_time: if True, return expire_time in tuple
            (default False)
        :param bool tag: if True, return tag in tuple (default False)
        :param bool retry: retry if database timeout occurs (default False)
        :return: key and value item pair
        :raises KeyError: if cache is empty
        :raises Timeout: if database timeout occurs

        """
        order = ('ASC', 'DESC')
        select = (
            'SELECT rowid, key, expire_time, tag, value'
            f' FROM "{self.table}" ORDER BY rowid %s LIMIT 1'
        ) % order[last]

        while True:
            while True:
                with self._transact(retry) as sql:
                    rows = sql(select).fetchall()
                    if not rows: raise KeyError('dictionary is empty')
                    (
                        (
                            rowid,
                            key,
                            db_expire,
                            db_tag,
                            value,
                        ),
                    ) = rows

                    if db_expire is not None and db_expire < time.time():
                        sql(f'DELETE FROM "{self.table}" WHERE rowid = ?', (rowid,))
                    else:
                        break

            break

        if expire_time and tag: return (key, value), db_expire, db_tag
        elif expire_time: return (key, value), db_expire
        elif tag: return (key, value), db_tag
        return key, value
    
    async def apeekitem(self, last: bool = True, expire_time: bool = False, tag: bool = False, retry: bool = False):
        """Peek at key and value item pair in cache based on iteration order.

        Expired items are deleted from cache. Operation is atomic. Concurrent
        operations will be serialized.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        >>> cache = Cache()
        >>> for num, letter in enumerate('abc'):
        ...     cache[letter] = num
        >>> cache.peekitem()
        ('c', 2)
        >>> cache.peekitem(last=False)
        ('a', 0)

        :param bool last: last item in iteration order (default True)
        :param bool expire_time: if True, return expire_time in tuple
            (default False)
        :param bool tag: if True, return tag in tuple (default False)
        :param bool retry: retry if database timeout occurs (default False)
        :return: key and value item pair
        :raises KeyError: if cache is empty
        :raises Timeout: if database timeout occurs

        """
        order = ('ASC', 'DESC')
        select = (
            'SELECT rowid, key, expire_time, tag, value'
            f' FROM "{self.table}" ORDER BY rowid %s LIMIT 1'
        ) % order[last]

        while True:
            while True:
                async with self._atransact(retry) as sql:
                    rows = await (await sql(select)).fetchall()
                    if not rows: raise KeyError('dictionary is empty')
                    (
                        (
                            rowid,
                            key,
                            db_expire,
                            db_tag,
                            value,
                        ),
                    ) = rows
                    if db_expire is not None and db_expire < time.time():
                        await sql(f'DELETE FROM "{self.table}" WHERE rowid = ?', (rowid,))
                    else:
                        break
            break

        if expire_time and tag: return (key, value), db_expire, db_tag
        elif expire_time: return (key, value), db_expire
        elif tag: return (key, value), db_tag
        return key, value

    def evict(self, tag: str, retry: bool = False):
        """Remove items with matching `tag` from cache.

        Removing items is an iterative process. In each iteration, a subset of
        items is removed. Concurrent writes may occur between iterations.

        If a :exc:`Timeout` occurs, the first element of the exception's
        `args` attribute will be the number of items removed before the
        exception occurred.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param str tag: tag identifying items
        :param bool retry: retry if database timeout occurs (default False)
        :return: count of rows removed
        :raises Timeout: if database timeout occurs

        """
        select = (
            f'SELECT rowid FROM "{self.table}"'
            ' WHERE tag = ? AND rowid > ?'
            ' ORDER BY rowid LIMIT ?'
        )
        args = [tag, 0, 100]
        return self._select_delete(select, args, arg_index=1, retry=retry)
    

    async def aevict(self, tag: str, retry: bool = False):
        """Remove items with matching `tag` from cache.

        Removing items is an iterative process. In each iteration, a subset of
        items is removed. Concurrent writes may occur between iterations.

        If a :exc:`Timeout` occurs, the first element of the exception's
        `args` attribute will be the number of items removed before the
        exception occurred.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param str tag: tag identifying items
        :param bool retry: retry if database timeout occurs (default False)
        :return: count of rows removed
        :raises Timeout: if database timeout occurs

        """
        select = (
            f'SELECT rowid FROM "{self.table}"'
            ' WHERE tag = ? AND rowid > ?'
            ' ORDER BY rowid LIMIT ?'
        )
        args = [tag, 0, 100]
        return await self._aselect_delete(select, args, arg_index=1, retry=retry)
    

    def clear(self, retry=False):
        """Remove all items from cache.

        Removing items is an iterative process. In each iteration, a subset of
        items is removed. Concurrent writes may occur between iterations.

        If a :exc:`Timeout` occurs, the first element of the exception's
        `args` attribute will be the number of items removed before the
        exception occurred.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param bool retry: retry if database timeout occurs (default False)
        :return: count of rows removed
        :raises Timeout: if database timeout occurs

        """
        select = (
            f'SELECT rowid FROM "{self.table}"'
            ' WHERE rowid > ?'
            ' ORDER BY rowid LIMIT ?'
        )
        args = [0, 100]
        return self._select_delete(select, args, retry=retry)


    async def aclear(self, retry=False):
        """Remove all items from cache.

        Removing items is an iterative process. In each iteration, a subset of
        items is removed. Concurrent writes may occur between iterations.

        If a :exc:`Timeout` occurs, the first element of the exception's
        `args` attribute will be the number of items removed before the
        exception occurred.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param bool retry: retry if database timeout occurs (default False)
        :return: count of rows removed
        :raises Timeout: if database timeout occurs

        """
        select = (
            f'SELECT rowid FROM "{self.table}"'
            ' WHERE rowid > ?'
            ' ORDER BY rowid LIMIT ?'
        )
        args = [0, 100]
        return await self._aselect_delete(select, args, retry=retry)

    def expire(self, now: Optional[float] = None, retry: bool = False):
        """Remove expired items from cache.

        Removing items is an iterative process. In each iteration, a subset of
        items is removed. Concurrent writes may occur between iterations.

        If a :exc:`Timeout` occurs, the first element of the exception's
        `args` attribute will be the number of items removed before the
        exception occurred.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param float now: current time (default None, ``time.time()`` used)
        :param bool retry: retry if database timeout occurs (default False)
        :return: count of items removed
        :raises Timeout: if database timeout occurs

        """
        select = (
            f'SELECT rowid, expire_time FROM "{self.table}"'
            ' WHERE ? < expire_time AND expire_time < ?'
            ' ORDER BY expire_time LIMIT ?'
        )
        args = [0, now or time.time(), 100]
        return self._select_delete(select, args, row_index=1, retry=retry)

    
    async def aexpire(self, now: Optional[float] = None, retry: bool = False):
        """Remove expired items from cache.

        Removing items is an iterative process. In each iteration, a subset of
        items is removed. Concurrent writes may occur between iterations.

        If a :exc:`Timeout` occurs, the first element of the exception's
        `args` attribute will be the number of items removed before the
        exception occurred.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param float now: current time (default None, ``time.time()`` used)
        :param bool retry: retry if database timeout occurs (default False)
        :return: count of items removed
        :raises Timeout: if database timeout occurs

        """
        select = (
            f'SELECT rowid, expire_time FROM "{self.table}"'
            ' WHERE ? < expire_time AND expire_time < ?'
            ' ORDER BY expire_time LIMIT ?'
        )
        args = [0, now or time.time(), 100]
        return await self._aselect_delete(select, args, row_index=1, retry=retry)
    
    
    def cull(self, retry: bool = False):
        """Cull items from cache until volume is less than size limit.

        Removing items is an iterative process. In each iteration, a subset of
        items is removed. Concurrent writes may occur between iterations.

        If a :exc:`Timeout` occurs, the first element of the exception's
        `args` attribute will be the number of items removed before the
        exception occurred.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param bool retry: retry if database timeout occurs (default False)
        :return: count of items removed
        :raises Timeout: if database timeout occurs

        """
        now = time.time()
        # Remove expired items.
        count = self.expire(now)
        # Remove items by policy.
        select_policy = self.eviction_policies['cull']
        if select_policy is None: return 0

        select_key = select_policy.format(fields='key', now=now)
        # select_filename = select_policy.format(fields='filename', now=now)

        try:
            while self.volume() > self.size_limit:
                with self._transact(retry) as sql:
                    rows = sql(select_key, (10,)).fetchall()
                # with self._transact(retry) as (sql, cleanup):
                #     rows = sql(select_filename, (10,)).fetchall()

                    if not rows: break
                    count += len(rows)
                    delete = (
                        f'DELETE FROM "{self.table}" WHERE rowid IN (%s)'
                        % select_policy.format(fields='rowid', now=now)
                    )
                    sql(delete, (10,))

                    # for (filename,) in rows:
                    #     cleanup(filename)
        except Timeout:
            raise Timeout(count) from None

        return count
    
    async def acull(self, retry: bool = False):
        """Cull items from cache until volume is less than size limit.

        Removing items is an iterative process. In each iteration, a subset of
        items is removed. Concurrent writes may occur between iterations.

        If a :exc:`Timeout` occurs, the first element of the exception's
        `args` attribute will be the number of items removed before the
        exception occurred.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default).

        :param bool retry: retry if database timeout occurs (default False)
        :return: count of items removed
        :raises Timeout: if database timeout occurs

        """
        now = time.time()
        # Remove expired items.
        count = await self.aexpire(now)
        # Remove items by policy.
        select_policy = self.eviction_policies['cull']
        if select_policy is None: return 0

        select_key = select_policy.format(fields='key', now=now)
        try:
            while (await self.avolume()) > self.size_limit:
                async with self._atransact(retry) as sql:
                    rows = await (await sql(select_key, (10,))).fetchall()
                    if not rows: break
                    count += len(rows)
                    delete = (
                        f'DELETE FROM "{self.table}" WHERE rowid IN (%s)'
                        % select_policy.format(fields='rowid', now=now)
                    )
                    await sql(delete, (10,))
        except Timeout:
            raise Timeout(count) from None
        return count


    """
    Class Methods
    """

    def __setitem__(self, key: str, value: t.Any):
        """Set corresponding `value` for `key` in cache.

        :param key: key for item
        :param value: value for item
        :return: corresponding value
        :raises KeyError: if key is not found

        """
        self.set(key, value, retry=True)
    

    def __getitem__(self, key: str):
        """Return corresponding value for `key` from cache.

        :param key: key matching item
        :return: corresponding value
        :raises KeyError: if key is not found

        """
        value = self.get(key, default=ENOVAL, retry=True)
        if value is ENOVAL: raise KeyError(key)
        return value
    

    def __contains__(self, key: str) -> bool:
        """Return `True` if `key` matching item is found in cache.

        :param key: key matching item
        :return: True if key matching item

        """
        return self.contains(key)
    
    def __delitem__(self, key: str, retry: bool = True):
        """Delete corresponding item for `key` from cache.

        Raises :exc:`Timeout` error when database timeout occurs and `retry` is
        `False` (default `True`).

        :param key: key matching item
        :param bool retry: retry if database timeout occurs (default True)
        :raises KeyError: if key is not found
        :raises Timeout: if database timeout occurs

        """
        return self._delete_key_(key, retry=retry)
    

    def __len__(self):
        """Count of items in cache including expired items."""
        return self.length()
    



from __future__ import annotations

"""
LazyOps Commands

A collection of useful commands to streamline your workflow and improve productivity.
"""

import typer
from . import keygen

cmd = typer.Typer(no_args_is_help = True, help = "LZO CLI")
cmd.add_typer(keygen.cmd, name="kg", help = "Key Generation Commands")

def main():
    cmd()

if __name__ == '__main__':
    main()



from __future__ import annotations

"""
Sqlite Configuration
"""

from ..base.config import (
    SqliteDsn,
    BaseDBConfig,
    BaseDBSettings,
    DatabaseConfig,
)
from lzl.types import eproperty
import typing as t

class BaseSqliteConfig(BaseDBConfig):
    """
    The Base Sqlite Config
    """
    url: t.Optional[SqliteDsn] = None
    readonly_url: t.Optional[SqliteDsn] = None
    superuser_url: t.Optional[SqliteDsn] = None


    @classmethod
    def get_adapter_prefix(cls) -> str:
        """
        Returns the adapter prefix
        """
        return 'sqlite'
    

    @eproperty
    def cli_connect_string(self) -> str:
        """
        Returns the cli connect string

        export PGPASSWORD=<password> psql -h <host> -p <port> -U <user> -d <database>
        """
        raise NotImplementedError
    

    def get_cli_connect_string(
        self,
        superuser: t.Optional[bool] = None,
        readonly: t.Optional[bool] = None,
        database: t.Optional[str] = None,
        **kwargs,
    ) -> str:
        """
        Returns the cli connect string
        """
        raise NotImplementedError
    
    def get_cli_debug_args(self) -> str:
        """
        Returns the cli debug args
        """
        return ''
    
    def get_cli_safe_cmd(self, cmd: str) -> str:
        """
        Returns the cli safe cmd
        """
        raise NotImplementedError


class SqliteSettings(BaseSqliteConfig, BaseDBSettings):
    """
    The Sqlite Settings
    """

    class Config:
        env_prefix = "SQLITE_"
        case_sensitive = False


class SqliteConfig(BaseSqliteConfig, DatabaseConfig):
    """
    The Sqlite Config
    """
    pass






from __future__ import annotations

"""Adapter-specific configuration registries for LazyOps databases."""

import typing as t
from .base.config import BaseDBConfig, BaseDBSettings, DBConfigT, DBSettingsT
# from .pg.config import PostgresConfig, PostgresSettings
from .postgres.config import PostgresConfig, PostgresSettings
from .sqlite.config import SqliteConfig, SqliteSettings

ADAPTER_TO_CONFIG: t.Dict[str, t.Type[DBConfigT]] = {
    'postgres': PostgresConfig,
    'postgresql': PostgresConfig,

    'sqlite': SqliteConfig,
}
ADAPTER_TO_SETTINGS: t.Dict[str, t.Type[DBSettingsT]] = {
    'postgres': PostgresSettings,
    'postgresql': PostgresSettings,

    'sqlite': SqliteSettings,
}

DATABASE_CONFIG_CLASSES = t.Union[PostgresConfig, SqliteConfig]

__all__ = [
    "ADAPTER_TO_CONFIG",
    "ADAPTER_TO_SETTINGS",
    "DATABASE_CONFIG_CLASSES",
]



from __future__ import annotations

"""
Unified API Client based on `niquests`
"""
import typing as t
from .client import Client, DEFAULT_RETRIES, DEFAULT_TIMEOUT
from .api import (
    request,
    arequest,
    get,
    aget,
    put,
    aput,
    patch,
    apatch,
    head,
    ahead,
    options,
    aoptions,
    delete,
    adelete,
    post,
    apost,   
)

if t.TYPE_CHECKING:
    from niquests import Response, AsyncResponse
    from niquests.sessions import HttpAuthenticationType, HttpMethodType, QueryParameterType, BodyType, HeadersType, CookiesType, MultiPartFilesType, MultiPartFilesAltType, TimeoutType, ProxyType, HookType, TLSVerifyType, TLSClientCertType
    from niquests.sessions import PreparedRequest
    from niquests._async import RetryType
    from niquests._typing import AsyncHttpAuthenticationType, AsyncBodyType
    from niquests.structures import QuicSharedCache
    from niquests._async import AsyncBaseAdapter
    from niquests.exceptions import (
        ConnectionError,
        ConnectTimeout,
        FileModeWarning,
        HTTPError,
        JSONDecodeError,
        ReadTimeout,
        RequestException,
        RequestsDependencyWarning,
        Timeout,
        TooManyRedirects,
        URLRequired,
    )



"""
Kinde OAuth2 Provider
"""
from __future__ import annotations

from .config import KindeOAuth2Config
from .client import KindeOAuth2Client
from .token import (
    IDToken as KindeIDToken,
    AccessToken as KindeAccessToken,
    OAuth2Token as KindeOAuth2Token,
)
from .types import UserInfo as KindeUserInfo

from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    
    TokenT = Union[KindeAccessToken, KindeIDToken, KindeOAuth2Token]


import os
import sys
import anyio
import typing as t

# if 3.10
if sys.version_info.minor < 9:
    import pathlib
    from pathlib import PosixPath, WindowsPath, Path, PurePath, _ignore_error
    from pathlib import _NormalAccessor as NormalAccessor
    from pathlib import _make_selector as _sync_make_selector
    from pathlib import _PosixFlavour, _WindowsFlavour
    try:
        from pathlib import _getfinalpathname
        from aiopath.wrap import func_to_async_func
        _async_getfinalpathname = func_to_async_func(_getfinalpathname)
    except ImportError:
        def _getfinalpathname(*args, **kwargs):
            raise ImportError("_getfinalpathname() requires a Windows/NT platform")
    
        async def _async_getfinalpathname(*args, **kwargs):
            raise ImportError("_getfinalpathname() requires a Windows/NT platform")
else:
    from . import lib as pathlib
    from .lib import PosixPath, WindowsPath, Path, PurePath, _ignore_error
    from .lib import _NormalAccessor as NormalAccessor
    from .lib import _make_selector as _sync_make_selector
    from .lib import _PosixFlavour, _WindowsFlavour
    from .lib import _getfinalpathname, _async_getfinalpathname

from .lib import get_file_info, guess_type

TextMode = \
  t.Literal['r', 'w', 'a', 'x', 'r+', 'w+', 'a+', 'x+']
BinaryMode = \
  t.Literal['rb', 'wb', 'ab', 'xb', 'r+b', 'w+b', 'a+b', 'x+b']
FileMode = t.Union[TextMode, BinaryMode]


DEFAULT_ENCODING: t.Final[str] = 'utf-8'
ON_ERRORS: t.Final[str] = 'ignore'
NEWLINE: t.Final[str] = '\n'

BEGINNING: t.Final[int] = 0
CHUNK_SIZE: t.Final[int] = 4 * 1_024

SEP: t.Final[str] = '\n'
ENCODING: t.Final[str] = 'utf-8'
ERRORS: t.Final[str] = 'replace'

FileData = t.Union[bytes, str]
Handle = anyio.AsyncFile
Paths = t.Union[Path, os.PathLike, str]


from __future__ import annotations

"""
Cachify Client Mixin
"""

import os
import abc
import time
import asyncio
import subprocess
from lzl import load
from lzl.types import eproperty, Literal
from kvdb.io import cachify as _cachify
from .utils import create_cachify_build_name_func, get_user_agent, get_default_headers, http_retry_wrapper
from typing import Optional, Type, TypeVar, Union, Set, List, Any, Dict, Callable, Awaitable, overload, TYPE_CHECKING

if TYPE_CHECKING:
    from lzl.logging import Logger
    from lzo.types.base import RegisteredSettings, BaseModel
    from lzl.pool import ThreadPool
    from .base import DataSchema, ResultSchema

if load.TYPE_CHECKING:
    from lzl.api import aiohttpx
    from lzl.api.aiohttpx import Response, Client as HTTPClient

else:
    aiohttpx = load.lazy_load('lzl.api.aiohttpx', install_missing = False)


ResponseDataType = Literal['json', 'text', 'bytes']
ResponseType = Optional[Union[ResponseDataType, Type['ResultSchema'], Type['BaseModel'], str]]
ResponseT = Union[str, bytes, Dict[str, Any], List[Any], 'ResultSchema', List['ResultSchema'], Dict[str, 'ResultSchema'], List[Dict[str, 'ResultSchema']]]


cachify = _cachify.create_context()

class CachifyClientMixin(abc.ABC):
    """
    Cachify Client Mixin
    """

    cachify_enabled: Optional[bool] = True
    cachify_ttl: Optional[int] = None
    cachify_shared_global: Optional[bool] = False
    
    if TYPE_CHECKING:
        name: str
        settings: RegisteredSettings

        serialization: Optional[str] = 'json'
        serialization_compression: Optional[str] = 'zstd'
        serialization_compression_level: Optional[int] = 19 # 6 for default (lz4, zlib) 19 for zstd


    """
    Cachify Validation Methods
    """

    def cachify_validator_is_not_cachable(self, *args, cachable: Optional[bool] = True, **kwargs) -> bool:
        """
        Checks if the function is not cachable
        """
        from kvdb.io.cachify.helpers import is_not_cachable
        return is_not_cachable(*args, cachable = cachable, **kwargs)
    
    def cachify_validator_is_disabled(self, *args, disabled: Optional[bool] = None, **kwargs) -> bool:
        """
        Checks if the function is disabled
        """
        from kvdb.io.cachify.helpers import is_disabled
        return is_disabled(*args, disabled = disabled, **kwargs)
    
    def cachify_validator_is_overwrite(self, *args, overwrite: Optional[bool] = None, **kwargs) -> bool:
        """
        Checks if the function is overwrite
        """
        from kvdb.io.cachify.helpers import is_overwrite
        return is_overwrite(*args, overwrite = overwrite, **kwargs)
    
    def cachify_get_exclude_keys(self, func: str, **kwargs) -> List[str]:
        """
        Gets the exclude keys
        """
        return None
    
    def cachify_get_name_builder_kwargs(self, func: str, **kwargs) -> Dict[str, Any]:
        """
        Gets the name builder kwargs
        """
        return {}
    
    def cachify_get_extra_serialization_kwargs(self, func: str, **kwargs) -> Dict[str, Any]:
        """
        Gets the extra serialization kwargs
        """
        return {}

    def cachify_create_base_name(self, func: str, **kwargs) -> str:
        """
        Creates the base name
        """
        return f'{self.settings.ctx.module_name}.{self.name}' if self.cachify_shared_global  else \
            f'{self.settings.ctx.module_name}.{self.settings.app_env.name}.{self.name}'
    

    def validate_cachify(self, func: str, **kwargs) -> Dict[str, Any]:
        """
        Validates the cachify function
        """
        if not self.cachify_enabled: return None
        base_name = self.cachify_create_base_name(func, **kwargs)
        if 'name' not in kwargs: kwargs['name'] = create_cachify_build_name_func(
            base_name = base_name,
            **self.cachify_get_name_builder_kwargs(func, **kwargs),
        )
        if 'ttl' not in kwargs: kwargs['ttl'] = self.cachify_ttl
        if 'serializer' not in kwargs and 'encoder' not in kwargs and 'decoder' not in kwargs:
            kwargs['serializer'] = self.serialization
            kwargs['serializer_kwargs'] = {
                'compression': self.serialization_compression,
                'compression_level': self.serialization_compression_level,
                'raise_errors': True,
                **self.cachify_get_extra_serialization_kwargs(func, **kwargs),
            }
        if 'verbosity' not in kwargs and self.settings.is_local_env: kwargs['verbosity'] = 2
        kwargs['disabled'] = self.cachify_validator_is_not_cachable
        kwargs['overwrite_if'] = self.cachify_validator_is_overwrite
        kwargs['disabled_if'] = self.cachify_validator_is_disabled
        if exclude_keys := self.cachify_get_exclude_keys(func, **kwargs):
            kwargs['exclude_keys'] = exclude_keys
        kwargs['exclude_null'] = True
        return kwargs
    

class CachifyHTTPClientMixin(CachifyClientMixin):

    default_cachable: Optional[bool] = True


    def cachify_validator_is_not_cachable(self, *args, cachable: Optional[bool] = None, **kwargs) -> bool:
        """
        Checks if the function is not cachable
        """
        if cachable is None: cachable = self.default_cachable
        return super().cachify_validator_is_not_cachable(*args, cachable = cachable, **kwargs)


    def cachify_validator_is_disabled(self, *args, disable_cache: Optional[bool] = None, **kwargs) -> bool:
        """
        Checks if the function is disabled
        """
        return disable_cache
    
    def cachify_validator_is_overwrite(self, *args, overwrite_cache: Optional[bool] = None, **kwargs) -> bool:
        """
        Checks if the function is overwrite
        """
        return overwrite_cache

    def cachify_get_exclude_keys(self, func: str, **kwargs) -> List[str]:
        """
        Gets the exclude keys
        """
        return [
            'background',
            'callback',
            'retryable',
            'retry_limit',
            'validate_url',
            'cachable',
            'disable_cache',
            'overwrite_cache',
            'verify',
        ]
    
    def cachify_get_name_builder_kwargs(self, func: str, **kwargs) -> Dict[str, Any]:
        """
        Gets the name builder kwargs
        """
        return {
            'include_http_methods': True,
            'special_names': ['pdftotext', 'csx'],
        }



class BaseHTTPClientMixin(abc.ABC):
    """
    The Base HTTP Client Mixin
    """

    http_timeout: Optional[float] = None
    http_errors_verbose: Optional[bool] = True
    http_errors_truncate_length: Optional[int] = 1000

    default_endpoint: Optional[str] = None
    default_headers: Optional[Dict[str, Any]] = None
    default_response_type: ResponseType = 'json'

    default_retryable: Optional[bool] = False
    default_retryable_limit: Optional[int] = 3

    _extra: Dict[str, Any] = {}

    if TYPE_CHECKING:
        settings: RegisteredSettings
        pooler: ThreadPool
        schemas: Dict[str, Union[Type['DataSchema'], Type['ResultSchema']]]
        logger: Logger
        autologger: Logger


    @eproperty
    def endpoint(self) -> str:
        """
        Returns the endpoint

        - Override this to add custom endpoints
        """
        return self.default_endpoint
    
    @eproperty
    def headers(self) -> Dict[str, Any]:
        """
        Returns the headers

        - Override this to add custom headers
        """
        base_headers = get_default_headers()
        if self.default_headers: base_headers.update(self.default_headers)
        return base_headers
    
    @eproperty
    def http_client_kwargs(self) -> Dict[str, Any]:
        """
        Returns the http client kwargs

        - Override this to add custom kwargs
        """
        return {}
    
    @eproperty
    def http_client_timeout(self) -> Optional[float]:
        """
        Returns the http client timeout
        """
        if hasattr(self.settings, 'clients') and hasattr(self.settings.clients, 'http_pool'):
            return self.http_timeout or self.settings.clients.http_pool.default_timeout
        return self.http_timeout or 60
    
    @eproperty
    def http_client_limits(self) -> Dict[str, int]:
        """
        Returns the http client limits
        """
        if hasattr(self.settings, 'clients') and hasattr(self.settings.clients, 'http_pool'):
            return {
                'max_connections': self.settings.clients.http_pool.max_connections,
                'max_keepalive_connections': self.settings.clients.http_pool.max_keepalive_connections,
                'keepalive_expiry': self.settings.clients.http_pool.keepalive_expiry,
            }
    
    def wrap_retryable_method(
        self,
        func: Union[Callable[..., Any], str],
        retryable: Optional[bool] = None,
        retry_limit: Optional[int] = None,
    ) -> Union[Callable[..., 'Response'], Callable[..., Awaitable['Response']]]:
        """
        Wraps a retryable method
        """
        if isinstance(func, str): func = getattr(self.api, func)
        if retryable is None: retryable = self.default_retryable
        if not retryable: return func
        if retry_limit is None: retry_limit = self.default_retryable_limit
        return http_retry_wrapper(max_tries = retry_limit + 1)(func)
    
    def generate_headers(self, browser: Optional[str] = 'chrome', **kwargs) -> Dict[str, str]:
        """
        Generates the headers
        """
        return get_default_headers(browser = browser, **kwargs)

    def configure_http_client(self, *args, **kwargs) -> 'HTTPClient':
        """
        Configures the HTTP Client
        """
        from lzl.api import aiohttpx
        limits = aiohttpx.Limits(**self.http_client_limits)
        return aiohttpx.Client(
            base_url = self.endpoint,
            timeout = self.http_client_timeout,
            limits = limits,
            headers = self.headers,
            verify = False,
            **self.http_client_kwargs, 
        )


    @property
    def api(self) -> 'HTTPClient':
        """
        Returns the API Client
        """
        if self._extra.get('api') is None:
            self._extra['api'] = self.configure_http_client()
        return self._extra['api']
    
    def reset_api(self):
        """
        Resets the API Client
        """
        if self._extra.get('api') is not None:
            self._extra['api'].close()
            self._extra['api'] = None
    
    async def areset_api(self):
        """
        Resets the API Client
        """
        if self._extra.get('api') is not None:
            await self._extra['api'].aclose()
            self._extra['api'] = None

    """
    Basic HTTP Methods
    """

    def _validate_url(self, url: str) -> Union[bool, str]:
        """
        Quickly validates a URL
        """
        try:
            response = self.api.head(url, follow_redirects = True)
            try:
                response.raise_for_status()
                return True
            except Exception as e:
                return f'[{response.status_code}]: {response.text}'
        except Exception as e:
            return f'[{type(e)}]: {str(e)}'
        
    
    def _fetch_content_type(self, url: str) -> Optional[str]:
        """
        Fetches the content type
        """
        try:
            response = self.api.head(url, follow_redirects = True)
            return response.headers.get('content-type')
        except Exception as e:
            return None

    def _get_(self, url: str, **kwargs) -> 'Response':
        """
        Makes a GET request
        """
        return self.api.get(url, **kwargs)
    
    def _post_(self, url: str, **kwargs) -> 'Response':
        """
        Makes a POST request
        """
        return self.api.post(url, **kwargs)
    

    async def _avalidate_url(self, url: str) -> Union[bool, str]:
        """
        Quickly validates a URL
        """
        try:
            response = await self.api.async_head(url, follow_redirects = True)
            try:
                response.raise_for_status()
                return True
            except Exception as e:
                return f'[{response.status_code}] {response.text}'
        except Exception as e:
            return f'[{type(e)}]: {str(e)}'
        
    async def _afetch_content_type(self, url: str) -> Optional[str]:
        """
        Fetches the content type
        """
        try:
            response = await self.api.async_head(url, follow_redirects = True)
            return response.headers.get('content-type')
        except Exception as e:
            return None

    async def _aget_(self, url: str, **kwargs) -> 'Response':
        """
        Makes a GET request
        """
        return await self.api.async_get(url, **kwargs)
    
    async def _apost_(self, url: str, **kwargs) -> 'Response':
        """
        Makes a POST request
        """
        return await self.api.async_post(url, **kwargs)
    
    def handle_response(
        self, 
        response: 'Response', 
        response_type: ResponseType = None,
    ) -> ResponseT:
        """
        Handle the response
        """
        try:
            response.raise_for_status()
            if response_type is None: response_type = self.default_response_type
            if not response_type: return response
            if not isinstance(response_type, str):
                # Must be a model
                return response_type.model_validate_json(response.text)
                # return response_type.model_validate(response.json())
            response_type = str(response_type).lower()
            if response_type == 'raw': return response
            if response_type == 'json': return response.json()
            if response_type == 'text': return response.text
            if response_type == 'bytes': return response.content
            if response_type in self.schemas:
                return self.schemas[response_type].model_validate_json(response.text)
            else:
                raise ValueError(f'Invalid Response Type: {response_type}')
        except aiohttpx.HTTPStatusError as e:
            if not self.http_errors_verbose or self.settings.is_production_env:
                self.logger.warning(f'[{response.status_code} - {e.request.url}]')
            else:
                response_text = response.text
                if self.http_errors_truncate_length:
                    response_text = response_text[:self.http_errors_truncate_length]
                self.logger.error(f'[{response.status_code} - {e.request.url}] {response_text}')
        except Exception as e:
            if not self.http_errors_verbose:
                self.logger.error(f'[{type(e)}] {str(e)}')
            else:
                response_text = response.text
                if self.http_errors_truncate_length:
                    response_text = response_text[:self.http_errors_truncate_length]
                self.logger.trace(f'Error in response: {response_text}', e)
            raise e
    
    """
    Enhanced HTTP Methods
    """

    def _get(
        self,
        url: str,
        *args,
        response_type: ResponseType = None,
        retryable: Optional[bool] = None,
        retry_limit: Optional[int] = None,
        **kwargs
    ) -> ResponseT:
        """
        Creates a GET Request and returns a `ReturnTypeT`
        """
        get_func = self.wrap_retryable_method(self._get_, retryable = retryable, retry_limit = retry_limit)
        response = get_func(url, *args, **kwargs)
        return self.handle_response(response, response_type = response_type)
    
    def _post(
        self,
        url: str,
        *args,
        response_type: ResponseType = None,
        retryable: Optional[bool] = None,
        retry_limit: Optional[int] = None,
        **kwargs
    ) -> ResponseT:
        """
        Creates a POST Request and returns a `ReturnTypeT`
        """
        post_func = self.wrap_retryable_method(self._post_, retryable = retryable, retry_limit = retry_limit)
        response = post_func(url, *args, **kwargs)
        return self.handle_response(response, response_type = response_type)
    
    async def _aget(
        self,
        url: str,
        *args,
        response_type: ResponseType = None,
        retryable: Optional[bool] = None,
        retry_limit: Optional[int] = None,
        **kwargs
    ) -> ResponseT:
        """
        Creates a GET Request and returns a `ReturnTypeT`
        """
        get_func = self.wrap_retryable_method(self._aget_, retryable = retryable, retry_limit = retry_limit)
        response = await get_func(url, *args, **kwargs)
        return self.handle_response(response, response_type = response_type)
    
    async def _apost(
        self,
        url: str,
        *args,
        response_type: ResponseType = None,
        retryable: Optional[bool] = None,
        retry_limit: Optional[int] = None,
        **kwargs
    ) -> ResponseT:
        """
        Creates a POST Request and returns a `ReturnTypeT`
        """
        post_func = self.wrap_retryable_method(self._apost_, retryable = retryable, retry_limit = retry_limit)
        response = await post_func(url, *args, **kwargs)
        return self.handle_response(response, response_type = response_type)
    

    """
    Extended Functions
    """

    @eproperty
    def google_csx_base_url(self) -> Optional[str]:
        """
        Returns the Google CSX Base URL
        """
        return 'https://www.googleapis.com/customsearch/v1'
    

    @eproperty
    def google_csx_api_key(self) -> str:
        """
        Override this to add a Google CSX API Key
        """
        if hasattr(self.settings, 'clients') and hasattr(self.settings.clients, 'http_pool'):
            return getattr(self.settings.clients.http_pool, 'google_csx_api_key', None)
        return None
    
    @eproperty
    def google_csx_id(self) -> str:
        """
        Override this to add a Google CSX ID
        """
        if hasattr(self.settings, 'clients') and hasattr(self.settings.clients, 'http_pool'):
            return getattr(self.settings.clients.http_pool, 'google_csx_id', None)
        return None
    
    @eproperty
    def pdftotext_enabled(self) -> bool:
        """
        Returns whether pdftotext is enabled
        """
        try:
            subprocess.check_output(['which', 'pdftotext'])
            return True
        except Exception as e:
            return False
        
    @eproperty
    def pdftotext_cmd_args(self) -> str:
        """
        Returns the pdftotext command arguments
        """
        return '-layout -nopgbrk -eol unix -colspacing 0.7 -y 58 -x 0 -H 741 -W 596 - -'
        
    
    def _get_pdftotext_(
        self,
        url: str,
        validate_url: Optional[bool] = False,
        raise_errors: Optional[bool] = None,
        **kwargs
    ) -> Optional[str]:
        """
        Transform a PDF File to Text directly from URL
        """
        if validate_url:
            validate_result = self._validate_url(url)
            if validate_result != True:
                if raise_errors: raise ValueError(f'Invalid URL: {url}. {validate_result}')
                self.logger.error(f'Invalid URL: {url}. {validate_result}')
                return None
        
        cmd = f'curl -s {url} | pdftotext {self.pdftotext_cmd_args}'
        process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        try:
            stdout, stderr = process.communicate()
            stdout = stdout.decode('utf-8')
            return stdout
        except Exception as e:
            stderr = stderr.decode('utf-8')
            self.logger.error(f'Error in pdftotext: {stderr}: {e}')
            if raise_errors: raise e
            return None
        
    def _get_pdftotext(
        self,
        url: str,
        validate_url: Optional[bool] = False,
        retryable: Optional[bool] = None,
        retry_limit: Optional[int] = None,
        raise_errors: Optional[bool] = None,
        **kwargs
    ) -> Optional[str]:
        """
        Transform a PDF File to Text directly from URL
        """
        if not self.pdftotext_enabled:
            raise ValueError('pdftotext is not enabled. Please install pdftotext')
        get_func = self.wrap_retryable_method(self._get_pdftotext_, retryable = retryable, retry_limit = retry_limit)
        return get_func(url, validate_url = validate_url, raise_errors = raise_errors, **kwargs)

    async def _aget_pdftotext_(
        self,
        url: str,
        validate_url: Optional[bool] = False,
        raise_errors: Optional[bool] = None,
        **kwargs
    ) -> Optional[str]:
        """
        Transform a PDF File to Text directly from URL
        """
        if validate_url:
            validate_result = await self._avalidate_url(url)
            if validate_result != True:
                if raise_errors: raise ValueError(f'Invalid URL: {url}. {validate_result}')
                self.logger.error(f'Invalid URL: {url}. {validate_result}')
                return None
        cmd = f'curl -s {url} | pdftotext {self.pdftotext_cmd_args}'
        process = await asyncio.subprocess.create_subprocess_shell(cmd, shell=True, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE)
        try:
            stdout, stderr = await process.communicate()
            stdout = stdout.decode('utf-8')
            return stdout
        except Exception as e:
            stderr = stderr.decode('utf-8')
            self.logger.error(f'Error in pdftotext: {stderr}: {e}')
            if raise_errors: raise e
            return None
        
    async def _aget_pdftotext(
        self,
        url: str,
        validate_url: Optional[bool] = False,
        retryable: Optional[bool] = None,
        retry_limit: Optional[int] = None,
        raise_errors: Optional[bool] = None,
        **kwargs
    ) -> Optional[str]:
        """
        Transform a PDF File to Text directly from URL
        """
        if not self.pdftotext_enabled:
            raise ValueError('pdftotext is not enabled. Please install pdftotext')
        get_func = self.wrap_retryable_method(self._aget_pdftotext_, retryable = retryable, retry_limit = retry_limit)
        return await get_func(url, validate_url = validate_url, raise_errors = raise_errors, **kwargs)


    """
    Google CSX Methods
    """

    def _get_csx_(
        self,
        *args,
        response_type: ResponseType = 'json',
        retryable: Optional[bool] = None,
        retry_limit: Optional[int] = None,
        **kwargs
    ) -> ResponseT:
        """
        Returns the Google CSX Results
        """
        url = self.google_csx_base_url
        get_func = self.wrap_retryable_method(self._get_, retryable = retryable, retry_limit = retry_limit)
        response = get_func(url, *args, **kwargs)
        if response.status_code == 429:
            self.logger.warning(f'[{response.status_code}] Rate Limit Exceeded. Retrying after a bit', colored = True, prefix = 'GoogleCSX')    
            attempts = 1
            while response.status_code == 429:
                time.sleep(10.0 * attempts)
                response = get_func(url, *args, **kwargs)
                if response.status_code != 429: 
                    self.autologger.info(f'[{response.status_code}] Completed Retry after {attempts} Attempts', colored = True, prefix = 'GoogleCSX')
                    break
                attempts += 1
        return self.handle_response(response, response_type = response_type)

    async def _aget_csx_(
        self,
        *args,
        response_type: ResponseType = 'json',
        retryable: Optional[bool] = None,
        retry_limit: Optional[int] = None,
        **kwargs
    ) -> ResponseT:
        """
        Creates a GET Request and returns a `ReturnTypeT` for Google CSX to handle rate limiting
        """
        get_func = self.wrap_retryable_method(self._aget_, retryable = retryable, retry_limit = retry_limit)
        url = self.google_csx_base_url
        response = await get_func(url, *args, **kwargs)
        if response.status_code == 429:
            self.logger.warning(f'[{response.status_code}] Rate Limit Exceeded. Retrying after a bit', colored = True, prefix = 'GoogleCSX')    
            attempts = 1
            while response.status_code == 429:
                await asyncio.sleep(10.0 * attempts)
                response = await get_func(url, *args, **kwargs)
                if response.status_code != 429: 
                    self.autologger.info(f'[{response.status_code}] Completed Retry after {attempts} Attempts', colored = True, prefix = 'GoogleCSX')
                    break
                attempts += 1
        return self.handle_response(response, response_type = response_type)
    


"""
Browser Mixin
"""

if TYPE_CHECKING:
    from playwright.sync_api import (
        PlaywrightContextManager, 
        Playwright, 
        Browser as SyncBrowser,
        BrowserType as SyncBrowserType
    )
    from playwright.async_api import (
        PlaywrightContextManager as AsyncPlaywrightContextManager, 
        Playwright as AsyncPlaywright,
        Browser as AsyncBrowser,
        BrowserType as AsyncBrowserType
    )

class BrowserMixin(abc.ABC):
    """
    The Browser Client Mixin
    """

    initialized: bool = False
    default_browser_type: Optional[str] = 'chromium'
    
    _extra: Dict[str, Any] = {}

    if TYPE_CHECKING:
        settings: RegisteredSettings
        logger: Logger
        pooler: ThreadPool
    
    @eproperty
    def downloads_path(self) -> Optional[str]:
        """
        Returns the downloads path
        """
        if hasattr(self.settings, 'module_path'):
            return self.settings.module_path.joinpath('data').as_posix()
        return os.getcwd()

    @eproperty
    def browser_launch_config(self) -> Dict[str, Any]:
        """
        Returns the browser launch config
        
        - Can be overridden by subclasses
        """
        return {
            'headless': True,
            'downloads_path': self.downloads_path,
        }


    @eproperty
    def initializing(self) -> bool:
        """
        Returns whether the client is initializing
        """
        return self._extra.get('initializing', False)
    
    @eproperty
    def cm(self) -> 'PlaywrightContextManager':
        """
        Returns the Playwright context manager
        """
        from playwright.sync_api import sync_playwright
        return sync_playwright()
    
    @eproperty
    def client(self) -> 'Playwright':
        """
        Returns the Playwright instance
        """
        return self.cm.start()
    
    @eproperty
    def browser(self) -> 'SyncBrowser':
        """
        Returns the browser instance
        """
        browser_t: 'SyncBrowserType' = getattr(self.client, self.default_browser_type)
        return browser_t.launch(**self.browser_launch_config)

    
    """
    Async Methods
    """

    @eproperty
    def acm(self) -> 'AsyncPlaywrightContextManager':
        """
        Returns the Playwright async context manager
        """
        from playwright.async_api import async_playwright
        return async_playwright()
    
    @property
    def aclient(self) -> 'AsyncPlaywright':
        """
        Returns the Playwright instance
        """
        if 'aclient' not in self._extra:
            raise RuntimeError('AsyncPlaywright not initialized. Initialize with Browser.ainit()')
        return self._extra['aclient']
    
    @aclient.setter
    def aclient(self, value: 'AsyncPlaywright'):
        """
        Sets the Playwright instance
        """
        self._extra['aclient'] = value
    
    @property
    def abrowser(self) -> 'AsyncBrowser':
        """
        Returns the browser instance
        """
        if 'abrowser' not in self._extra:
            raise RuntimeError('AsyncBrowser not initialized. Initialize with Browser.ainit()')
        return self._extra['abrowser']
    
    @abrowser.setter
    def abrowser(self, value: 'AsyncBrowser'):
        """
        Sets the browser instance
        """
        self._extra['abrowser'] = value


    def _set_aclient(self, task: asyncio.Task):
        """
        Sets the AsyncPlaywright instance
        """
        self.logger.info('Setting AsyncPlaywright')
        self.aclient = task.result()

    def _set_abrowser(self, task: asyncio.Task):
        """
        Sets the AsyncBrowser instance
        """
        self.logger.info('Setting AsyncBrowser')
        self.abrowser = task.result()

    async def ainit(self):
        """
        Initializes the browser
        """
        if self.initialized: return
        self.logger.info('Initializing Browser')
        self.initializing = True
        if 'aclient' not in self._extra:
            self.aclient = await self.acm.start()
        if 'abrowser' not in self._extra:
            browser_t: 'AsyncBrowserType' = getattr(self.aclient, self.default_browser_type)
            self.abrowser = await browser_t.launch(**self.browser_launch_config)
        self.initialized = True
        self.initializing = False
        return self
    

    def close(self):
        """
        Closes the browser
        """
        if 'abrowser' in self._extra:
            self.pooler.create_background(self.abrowser.close)
            _ = self._extra.pop('abrowser')
        if 'browser' in self._extra:
            self.browser.close()
            _ = self._extra.pop('browser')
        if 'aclient' in self._extra:
            self.pooler.create_background(self.aclient.stop)
            _ = self._extra.pop('aclient')
        if 'client' in self._extra:
            self.client.stop()
            _ = self._extra.pop('client')
        _ = self._extra.pop('cm', None)
        _ = self._extra.pop('acm', None)
        self.initialized = False
    
    async def aclose(self):
        """
        Closes the browser
        """
        if 'abrowser' in self._extra:
            await self.abrowser.close()
            _ = self._extra.pop('abrowser')
        if 'browser' in self._extra:
            self.browser.close()
            _ = self._extra.pop('browser')
        if 'aclient' in self._extra:
            await self.aclient.stop()
            _ = self._extra.pop('aclient')
        if 'client' in self._extra:
            await self.client.stop()
            _ = self._extra.pop('client')
        _ = self._extra.pop('cm', None)
        _ = self._extra.pop('acm', None)
        self.initialized = False
    


from __future__ import annotations

"""
Base Token Types
"""

import time
from lzo.types import BaseModel, Field, eproperty, model_validator
from typing import Optional, List, Dict, Any, Union, TYPE_CHECKING


class AccessTokenPayload(BaseModel):
    iss: Optional[str] = None
    sub: Optional[str] = None
    aud: Optional[Union[List, str]] = None
    exp: Optional[int] = None
    iat: Optional[int] = None
    nbf: Optional[int] = None
    jti: Optional[str] = None
    typ: Optional[str] = None
    azp: Optional[str] = None
    scope: Optional[str] = None

    value: Optional[str] = None # The actual token value
    

    def update(self, data: Dict[str, Any]):
        """
        Updates the payload
        """
        if not data.get('iat') or \
            data['iat'] < self.iat: return
        for key, value in data.items():
            if value and hasattr(self, key): setattr(self, key, value)
    
    @property
    def is_expired(self) -> bool:
        """
        Returns True if the Token is Expired
        """
        return self.exp < int(time.time()) if self.exp else False


class IDTokenPayload(AccessTokenPayload):
    hd: Optional[str] = None
    email: Optional[str] = None
    email_verified: Optional[bool] = None
    at_hash: Optional[str] = None
    name: Optional[str] = None
    picture: Optional[str] = None
    given_name: Optional[str] = None
    family_name: Optional[str] = None


class OAuth2Token(BaseModel):
    
    expires_in: Optional[int] = None
    refresh_token: Optional[str] = None
    scope: Optional[str] = None
    token_type: Optional[str] = None

    access_token: Optional[Union[AccessTokenPayload, str]] = None
    id_token: Optional[Union[IDTokenPayload, str]] = None
    issued_at: Optional[int] = None
    expiration_ts: Optional[float] = None
    provider: Optional[str] = None
    # value: Optional[str] = None # The actual token value

    @model_validator(mode = 'after')
    def validate_oauth_token(self):
        """
        Sets the Expires At
        """
        if self.expiration_ts is None:
            if self.expires_in is not None:
                self.expiration_ts = time.time() + (self.expires_in - 10)
            elif self.issued_at is not None:
                self.expiration_ts = self.issued_at + (self.expires_in - 10)
        elif self.expires_in is not None:
            self.expires_in = int(self.expiration_ts - time.time())
        return self
    

    @property
    def is_expired(self) -> bool:
        """
        Returns True if the Token is Expired
        """
        return self.expiration_ts < time.time() if self.expiration_ts else False
    

    def set_access_token(
        self, 
        data: Dict[str, Any],
        token: Optional[str] = None,
    ) -> None:
        """
        Sets the Access Token
        """
        access_token = AccessTokenPayload(**data)
        if token is not None: access_token.value = token
        elif isinstance(self.access_token, str):
            access_token.value = self.access_token
        self.access_token = access_token

    def set_id_token(
        self, 
        data: Dict[str, Any],
        token: Optional[str] = None,
    ) -> None:
        """
        Sets the ID Token
        """
        id_token = IDTokenPayload(**data)
        if token is not None: id_token.value = token
        elif isinstance(self.id_token, str):
            id_token.value = self.id_token
        self.id_token = id_token
    
    @property
    def identity(self) -> Optional[str]:
        """
        Returns the Identity
        """
        if 'identity' not in self._extra:
            if self.id_token and not isinstance(self.id_token, str) and self.id_token.sub:
                self._extra['identity'] = self.id_token.sub
            elif self.access_token and not isinstance(self.access_token, str) and self.access_token.sub:
                self._extra['identity'] = self.access_token.sub
        return self._extra.get('identity')

    @property
    def email(self) -> Optional[str]:
        """
        Returns the Email
        """
        if 'email' not in self._extra:
            if self.id_token and not isinstance(self.id_token, str) and self.id_token.email:
                self._extra['email'] = self.id_token.email
            elif self.access_token and not isinstance(self.access_token, str) and self.access_token.get('email'):
                self._extra['email'] = self.access_token.get('email')
        return self._extra.get('email')

    @property
    def roles(self) -> Optional[List[str]]:
        """
        Returns the Roles
        """
        if 'roles' not in self._extra:
            if self.id_token and not isinstance(self.id_token, str) and self.id_token.get('roles'):
                self._extra['roles'] = self.id_token.get('roles')
            elif self.access_token and not isinstance(self.access_token, str) and self.access_token.get('roles'):
                self._extra['roles'] = self.access_token.get('roles')
            
            if self._extra.get('roles') and isinstance(self._extra['roles'], list) and \
                isinstance(self._extra['roles'][0], dict):
                self._extra['roles'] = [role['name'] for role in self._extra['roles']]
        return self._extra.get('roles')

    @property
    def role(self) -> Optional[str]:
        """
        Returns the Role
        """
        if 'role' not in self._extra:
            if self.id_token and not isinstance(self.id_token, str) and self.id_token.get('role'):
                self._extra['role'] = self.id_token.get('role')
            elif self.access_token and not isinstance(self.access_token, str) and self.access_token.get('role'):
                self._extra['role'] = self.access_token.get('role')
            elif self.roles:
                self._extra['role'] = self.roles[0]
        return self._extra.get('role')
    
    @property
    def scopes(self) -> Optional[List[str]]:
        """
        Returns the Scopes
        """
        return ' '.split(self.scope) if self.scope else []
    
    @property
    def value(self) -> Optional[str]:
        """
        Returns the Value
        """
        if self.access_token and not isinstance(self.access_token, str) \
            and self.access_token.value:
            return self.access_token.value
        if self.id_token and not isinstance(self.id_token, str) \
            and self.id_token.value:
            return self.id_token.value
        return self.access_token
    
    @property
    def display_name(self) -> Optional[str]:
        """
        Returns the Display Name
        """
        if 'display_name' not in self._extra:
            if self.id_token and self.id_token.get('name'):
                self._extra['display_name'] = self.id_token.get('name')
            elif self.access_token and self.access_token.get('name'):
                self._extra['display_name'] = self.access_token.get('name')
        return self._extra.get('display_name')


    def merge(self, other: Dict[str, Any]):
        """
        Merges from a token payload
        """
        # This is likely a id_token
        if other.get('email'): 
            # print('Merging ID Token')
            self.id_token.update(other)
        elif other.get('sub'): 
            # print('Merging Access Token')
            self.access_token.update(other)
        return self
